import base64
import logging
import re
import subprocess
import os

import boto3
from botocore.exceptions import ClientError

from bentoml.exceptions import (
    BentoMLException,
    MissingDependencyException,
    AWSServiceError,
)

logger = logging.getLogger(__name__)

# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/\
# using-cfn-describing-stacks.html
FAILED_CLOUDFORMATION_STACK_STATUS = [
    "CREATE_FAILED",
    # Ongoing creation of one or more stacks with an expected StackId
    # but without any templates or resources.
    "REVIEW_IN_PROGRESS",
    "ROLLBACK_FAILED",
    # This status exists only after a failed stack creation.
    "ROLLBACK_COMPLETE",
    # Ongoing removal of one or more stacks after a failed stack
    # creation or after an explicitly cancelled stack creation.
    "ROLLBACK_IN_PROGRESS",
]

SUCCESS_CLOUDFORMATION_STACK_STATUS = ["CREATE_COMPLETE", "UPDATE_COMPLETE"]


def generate_aws_compatible_string(*items, max_length=63):
    """
    Generate a AWS resource name that is composed from list of string items. This
    function replaces all invalid characters in the given items into '-', and allow user
    to specify the max_length for each part separately by passing the item and its max
    length in a tuple, e.g.:

    >> generate_aws_compatible_string("abc", "def")
    >> 'abc-def'  # concatenate multiple parts

    >> generate_aws_compatible_string("abc_def")
    >> 'abc-def'  # replace invalid chars to '-'

    >> generate_aws_compatible_string(("ab", 1), ("bcd", 2), max_length=4)
    >> 'a-bc'  # trim based on max_length of each part
    """
    trimmed_items = [
        item[0][: item[1]] if type(item) == tuple else item for item in items
    ]
    items = [item[0] if type(item) == tuple else item for item in items]

    for i in range(len(trimmed_items)):
        if len("-".join(items)) <= max_length:
            break
        else:
            items[i] = trimmed_items[i]

    name = "-".join(items)
    if len(name) > max_length:
        raise BentoMLException(
            "AWS resource name {} exceeds maximum length of {}".format(name, max_length)
        )
    invalid_chars = re.compile("[^a-zA-Z0-9-]|_")
    name = re.sub(invalid_chars, "-", name)
    return name


def get_default_aws_region():
    try:
        aws_session = boto3.session.Session()
        region = aws_session.region_name
        if not region:
            return ""
        return aws_session.region_name
    except ClientError as e:
        # We will do nothing, if there isn't a default region
        logger.error("Encounter error when getting default region for AWS: %s", str(e))
        return ""


def ensure_sam_available_or_raise():
    try:
        import samcli

        if samcli.__version__ != "0.33.1":
            raise BentoMLException(
                "aws-sam-cli package requires version 0.33.1 "
                "Install the package with `pip install -U aws-sam-cli==0.33.1`"
            )
    except ImportError:
        raise MissingDependencyException(
            "aws-sam-cli package is required. Install "
            "with `pip install --user aws-sam-cli`"
        )


def call_sam_command(command, project_dir, region):
    command = ["sam"] + command

    # We are passing region as part of the param, due to sam cli is not currently
    # using the region that passed in each command.  Set the region param as
    # AWS_DEFAULT_REGION for the subprocess call
    logger.debug('Setting envar "AWS_DEFAULT_REGION" to %s for subprocess call', region)
    copied_env = os.environ.copy()
    copied_env["AWS_DEFAULT_REGION"] = region

    proc = subprocess.Popen(
        command,
        cwd=project_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=copied_env,
    )
    stdout, stderr = proc.communicate()
    logger.debug("SAM cmd %s output: %s", command, stdout.decode("utf-8"))
    return proc.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


def validate_sam_template(template_file, aws_region, sam_project_path):
    status_code, stdout, stderr = call_sam_command(
        ["validate", "--template-file", template_file, "--region", aws_region],
        project_dir=sam_project_path,
        region=aws_region,
    )
    if status_code != 0:
        error_message = stderr
        if not error_message:
            error_message = stdout
        raise BentoMLException(
            "Failed to validate lambda template. {}".format(error_message)
        )


def cleanup_s3_bucket_if_exist(bucket_name, region):
    s3_client = boto3.client('s3', region)
    s3 = boto3.resource('s3')
    try:
        logger.debug('Removing all objects inside bucket %s', bucket_name)
        s3.Bucket(bucket_name).objects.all().delete()
        logger.debug('Deleting bucket %s', bucket_name)
        s3_client.delete_bucket(Bucket=bucket_name)
    except ClientError as e:
        if e.response and e.response['Error']['Code'] == 'NoSuchBucket':
            # If there is no bucket, we just let it silently fail, dont have to do
            # any thing
            return
        else:
            raise e


def delete_cloudformation_stack(stack_name, region):
    cf_client = boto3.client("cloudformation", region)
    cf_client.delete_stack(StackName=stack_name)


def delete_ecr_repository(repository_name, region):
    try:
        ecr_client = boto3.client("ecr", region)
        ecr_client.delete_repository(repositoryName=repository_name, force=True)
    except ClientError as e:
        if e.response and e.response['Error']['Code'] == 'RepositoryNotFoundException':
            # Don't raise error, if the repo can't be found
            return
        else:
            raise e


def get_instance_public_ip(instance_id, region):
    ec2_client = boto3.client("ec2", region)
    response = ec2_client.describe_instances(InstanceIds=[instance_id])
    all_instances = response["Reservations"][0]["Instances"]
    if all_instances:
        if "PublicIpAddress" in all_instances[0]:
            return all_instances[0]["PublicIpAddress"]
    return ""


def get_instance_ip_from_scaling_group(autoscaling_group_names, region):
    asg_client = boto3.client("autoscaling", region)
    response = asg_client.describe_auto_scaling_groups(
        AutoScalingGroupNames=autoscaling_group_names
    )
    all_autoscaling_group_info = response["AutoScalingGroups"]

    all_instances = []
    if all_autoscaling_group_info:
        for group in all_autoscaling_group_info:
            for instance in group["Instances"]:
                endpoint = get_instance_public_ip(instance["InstanceId"], region)
                all_instances.append(
                    {
                        "instance_id": instance["InstanceId"],
                        "endpoint": endpoint,
                        "state": instance["LifecycleState"],
                        "health_status": instance["HealthStatus"],
                    }
                )

    return all_instances


def get_aws_user_id():
    return boto3.client("sts").get_caller_identity().get("Account")


def create_ecr_repository_if_not_exists(region, repository_name):
    ecr_client = boto3.client("ecr", region)
    try:
        result = ecr_client.describe_repositories(repositoryNames=[repository_name])
        repository_id = result['repositories'][0]['registryId']
    except ecr_client.exceptions.RepositoryNotFoundException:
        result = ecr_client.create_repository(repositoryName=repository_name)
        repository_id = result['repository']['registryId']
    return repository_id


def get_ecr_login_info(region, repository_id):
    ecr_client = boto3.client('ecr', region)
    token = ecr_client.get_authorization_token(registryIds=[repository_id])
    logger.debug("Getting docker login info from AWS")
    username, password = (
        base64.b64decode(token["authorizationData"][0]["authorizationToken"])
        .decode("utf-8")
        .split(":")
    )
    registry_url = token["authorizationData"][0]["proxyEndpoint"]

    return registry_url, username, password


def generate_bentoml_exception_from_aws_client_error(e, message_prefix=None):
    """parse botocore.exceptions.ClientError into Bento StatusProto

     We handle two most common errors when deploying to Sagemaker.
         1. Authentication issue/invalid access(InvalidSignatureException)
         2. resources not found (ValidationException)
     It will return correlated StatusProto(NOT_FOUND, UNAUTHENTICATED)

     Args:
         e: ClientError from botocore.exceptions
         message_prefix: string
     Returns:
         StatusProto
     """
    error_response = e.response.get("Error", {})
    error_code = error_response.get("Code", "Unknown")
    error_message = error_response.get("Message", "Unknown")
    error_log_message = (
        f"AWS ClientError - operation: {e.operation_name}, "
        f"code: {error_code}, message: {error_message}"
    )
    if message_prefix:
        error_log_message = f"{message_prefix}; {error_log_message}"
    logger.error(error_log_message)
    return AWSServiceError(error_log_message)


def describe_cloudformation_stack(region, stack_name):
    cf_client = boto3.client("cloudformation", region)
    try:
        cloudformation_stack_result = cf_client.describe_stacks(StackName=stack_name)
        stack_info = cloudformation_stack_result.get('Stacks')
        if len(stack_info) < 1:
            raise BentoMLException(f'Cloudformation {stack_name} not found')
        if len(stack_info) > 1:
            raise BentoMLException(
                f'Found more than one cloudformation stack for {stack_name}'
            )
        return stack_info[0]
    except ClientError as error:
        raise BentoMLException(
            f'Failed to describe CloudFormation {stack_name} {error}'
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import itertools
import json
from typing import Iterable, Iterator, Mapping

from bentoml.exceptions import BadInput
from bentoml.utils import catch_exceptions
from bentoml.utils.csv import (
    csv_quote,
    csv_row,
    csv_split,
    csv_splitlines,
    csv_unquote,
)
from bentoml.utils.lazy_loader import LazyLoader

pandas = LazyLoader('pandas', globals(), 'pandas')


def check_dataframe_column_contains(required_column_names, df):
    df_columns = set(map(str, df.columns))
    for col in required_column_names:
        if col not in df_columns:
            raise BadInput(
                "Missing columns: {}, required_column:{}".format(
                    ",".join(set(required_column_names) - df_columns), df_columns
                )
            )


@catch_exceptions(Exception, fallback=None)
def guess_orient(table, strict=False):
    if isinstance(table, list):
        if not table:
            if strict:
                return {"records", "values"}
            else:
                return "records"
        if isinstance(table[0], dict):
            return 'records'
        else:
            return 'values'
    elif isinstance(table, dict):
        if set(table) == {"columns", "index", "data"}:
            return 'split'
        if set(table) == {"schema", "data"} and "primaryKey" in table["schema"]:
            return 'table'
        if strict:
            return {'columns', "index"}
        else:
            return "columns"


class DataFrameState(object):
    def __init__(self, columns: Mapping[str, int] = None):
        self.columns = columns


def _from_json_records(state: DataFrameState, table: list):
    if state.columns is None:  # make header
        state.columns = {k: i for i, k in enumerate(table[0].keys())}
    for tr in table:
        yield csv_row(tr[c] for c in state.columns)


def _from_json_values(_: DataFrameState, table: list):
    for tr in table:
        yield csv_row(tr)


def _from_json_columns(state: DataFrameState, table: dict):
    if state.columns is None:  # make header
        state.columns = {k: i for i, k in enumerate(table.keys())}
    for row in next(iter(table.values())):
        yield csv_row(table[col][row] for col in state.columns)


def _from_json_index(state: DataFrameState, table: dict):
    if state.columns is None:  # make header
        state.columns = {k: i for i, k in enumerate(next(iter(table.values())).keys())}
        for row in table.keys():
            yield csv_row(td for td in table[row].values())
    else:
        for row in table.keys():
            yield csv_row(table[row][col] for col in state.columns)


def _from_json_split(state: DataFrameState, table: dict):
    table_columns = {k: i for i, k in enumerate(table['columns'])}

    if state.columns is None:  # make header
        state.columns = table_columns
        for row in table['data']:
            yield csv_row(row)
    else:
        idxs = [state.columns[k] for k in table_columns]
        for row in table['data']:
            yield csv_row(row[idx] for idx in idxs)


def _from_csv_without_index(state: DataFrameState, table: Iterator[str]):
    row_str = next(table)  # skip column names
    table_columns = tuple(csv_unquote(s) for s in csv_split(row_str, ','))

    if state.columns is None:
        state.columns = table_columns
        for row_str in table:
            if not row_str:  # skip blank line
                continue
            if not row_str.strip():
                yield csv_quote(row_str)
            else:
                yield row_str
    elif not all(
        c1 == c2 for c1, c2 in itertools.zip_longest(state.columns, table_columns)
    ):
        idxs = [state.columns[k] for k in table_columns]
        for row_str in table:
            if not row_str:  # skip blank line
                continue
            if not row_str.strip():
                yield csv_quote(row_str)
            else:
                tr = tuple(s for s in csv_split(row_str, ","))
                yield csv_row(tr[i] for i in idxs)
    else:
        for row_str in table:
            if not row_str:  # skip blank line
                continue
            if not row_str.strip():
                yield csv_quote(row_str)
            else:
                yield row_str


_ORIENT_MAP = {
    'records': _from_json_records,
    'columns': _from_json_columns,
    'values': _from_json_values,
    'split': _from_json_split,
    'index': _from_json_index,
    # 'table': _from_json_table,
}

PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS = {k for k in _ORIENT_MAP}


def _dataframe_csv_from_input(table: str, fmt, orient, state):
    try:
        if not fmt or fmt == "json":
            table = json.loads(table)
            if not orient:
                orient = guess_orient(table, strict=False)
            else:
                guessed_orient = guess_orient(table, strict=True)
                if orient != guessed_orient and orient not in guessed_orient:
                    print("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx", orient)
                    return None
            if orient not in _ORIENT_MAP:
                return None
            _from_json = _ORIENT_MAP[orient]
            try:
                return tuple(_from_json(state, table))
            except (TypeError, AttributeError, KeyError, IndexError):
                return None
        elif fmt == "csv":
            table = csv_splitlines(table)
            return tuple(_from_csv_without_index(state, table))
        else:
            return None
    except (json.JSONDecodeError):
        return None


def read_dataframes_from_json_n_csv(
    datas: Iterable[str],
    formats: Iterable[str],
    orient: str = None,
    columns=None,
    dtype=None,
) -> ("pandas.DataFrame", Iterable[slice]):
    '''
    load dataframes from multiple raw datas in json or csv format, efficiently

    Background: Each calling of pandas.read_csv or pandas.read_json cost about 100ms,
    no matter how many lines it contains. Concat jsons/csvs before read_json/read_csv
    to improve performance.
    '''
    state = DataFrameState(
        columns={k: i for i, k in enumerate(columns)} if columns else None
    )
    trs_list = tuple(
        _dataframe_csv_from_input(t, fmt, orient, state)
        for t, fmt in zip(datas, formats)
    )
    header = ",".join(csv_quote(td) for td in state.columns) if state.columns else None
    lens = tuple(len(trs) if trs else 0 for trs in trs_list)
    table = "\n".join(tr for trs in trs_list if trs is not None for tr in trs)
    try:
        if not header:
            df = pandas.read_csv(
                io.StringIO(table), index_col=None, dtype=dtype, header=None,
            )
        else:
            df = pandas.read_csv(
                io.StringIO("\n".join((header, table))), index_col=None, dtype=dtype,
            )
        return df, lens
    except pandas.errors.EmptyDataError:
        return None, lens



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import pathlib
from typing import Iterable, Iterator, Sequence, Tuple

from bentoml.adapters.base_input import BaseInputAdapter
from bentoml.adapters.utils import decompress_gzip_request
from bentoml.types import AwsLambdaEvent, FileLike, HTTPRequest, InferenceTask

ApiFuncArgs = Tuple[
    Sequence[FileLike],
]


class FileInput(BaseInputAdapter):
    """Convert incoming file data from http request, cli or lambda event into file
    stream object and pass down to user defined API functions

    Parameters
    ----------
    None

    Examples
    ----------
    Service using FileInput:

    .. code-block:: python

        import bentoml
        from PIL import Image
        import numpy as np

        from bentoml.frameworks.pytorch import PytorchModelArtifact
        from bentoml.adapters import FileInput

        FASHION_MNIST_CLASSES = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat',
                                 'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']

        @bentoml.env(pip_packages=['torch', 'pillow', 'numpy'])
        @bentoml.artifacts([PytorchModelArtifact('classifier')])
        class PyTorchFashionClassifier(bentoml.BentoService):

            @bentoml.api(input=FileInput(), batch=True)
            def predict(self, file_streams):
                img_arrays = []
                for fs in file_streams:
                    im = Image.open(fs).convert(mode="L").resize((28, 28))
                    img_array = np.array(im)
                    img_arrays.append(img_array)

                inputs = np.stack(img_arrays, axis=0)

                outputs = self.artifacts.classifier(inputs)
                return [FASHION_MNIST_CLASSES[c] for c in outputs]

    OR use FileInput with ``batch=False`` (the default):

    .. code-block:: python

        @bentoml.api(input=FileInput(), batch=False)
        def predict(self, file_stream):
            im = Image.open(file_stream).convert(mode="L").resize((28, 28))
            img_array = np.array(im)
            inputs = np.stack([img_array], axis=0)
            outputs = self.artifacts.classifier(inputs)
            return FASHION_MNIST_CLASSES[outputs[0]]

    Query with HTTP request performed by cURL::

        curl -i \\
          --header "Content-Type: image/jpeg" \\
          --request POST \\
          --data-binary @test.jpg \\
          localhost:5000/predict

    OR::

        curl -i \\
          -F image=@test.jpg \\
          localhost:5000/predict

    OR by an HTML form that sends multipart data:

    .. code-block:: html

        <form action="http://localhost:8000" method="POST"
              enctype="multipart/form-data">
            <input name="image" type="file">
            <input type="submit">
        </form>

    OR by python requests:

    .. code-block:: python

        import requests

        with open("test.jpg", "rb") as f:
            image_bytes = f.read()

        files = {
            "image": ("test.jpg", image_bytes),
        }
        response = requests.post(your_url, files=files)

    Query with CLI command::

        bentoml run PyTorchFashionClassifier:latest predict --input-file test.jpg

    OR infer all images under a folder with ten images each batch::

        bentoml run PyTorchFashionClassifier:latest predict \\
          --input-file folder/*.jpg --max-batch-size 10

    """

    HTTP_METHODS = ["POST"]
    BATCH_MODE_SUPPORTED = True

    @property
    def request_schema(self):
        return {
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {"file": {"type": "string", "format": "binary"}},
                }
            },
            "*/*": {"schema": {"type": "string", "format": "binary"}},
        }

    @decompress_gzip_request
    def from_http_request(self, req: HTTPRequest) -> InferenceTask[FileLike]:
        if req.headers.content_type == 'multipart/form-data':
            _, _, files = HTTPRequest.parse_form_data(req)
            if len(files) != 1:
                task = InferenceTask(data=None)
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} requires one and at"
                    " least one file at a time, if you just upgraded from"
                    " bentoml 0.7, you may need to use MultiFileAdapter instead",
                )
            else:
                input_file = next(iter(files.values()))
                task = InferenceTask(http_headers=req.headers, data=input_file)
        elif req.body:
            task = InferenceTask(
                http_headers=req.headers, data=FileLike(bytes_=req.body),
            )
        else:
            task = InferenceTask(data=None)
            task.discard(
                http_status=400,
                err_msg=f'BentoML#{self.__class__.__name__} unexpected HTTP request'
                ' format',
            )
        return task

    def from_aws_lambda_event(self, event: AwsLambdaEvent) -> InferenceTask[FileLike]:
        f = FileLike(bytes_=base64.decodebytes(event.get('body', "")))
        return InferenceTask(aws_lambda_event=event, data=f)

    def from_cli(self, cli_args: Tuple[str]) -> Iterator[InferenceTask[FileLike]]:
        import argparse

        parser = argparse.ArgumentParser()
        input_g = parser.add_mutually_exclusive_group(required=True)
        input_g.add_argument('--input', nargs="+", type=str)
        input_g.add_argument('--input-file', nargs="+")

        parsed_args, _ = parser.parse_known_args(list(cli_args))

        for t in self.from_inference_job(
            input_=parsed_args.input, input_file=parsed_args.input_file,
        ):
            t.cli_args = cli_args
            yield t

    def from_inference_job(  # pylint: disable=arguments-differ
        self, input_=None, input_file=None, **extra_args
    ) -> Iterator[InferenceTask[str]]:
        '''
        Generate InferenceTask from calling bentom_svc.run(input_=None, input_file=None)

        Parameters
        ----------
        input_ : str
            The input value

        input_file : str
            The URI/path of the input file

        extra_args : dict
            Additional parameters

        '''
        if input_file is not None:
            for d in input_file:
                uri = pathlib.Path(d).absolute().as_uri()
                yield InferenceTask(
                    inference_job_args=extra_args, data=FileLike(uri=uri)
                )
        else:
            for d in input_:
                yield InferenceTask(
                    inference_job_args=extra_args, data=FileLike(bytes_=d.encode()),
                )

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask[FileLike]]
    ) -> ApiFuncArgs:
        return (tuple(t.data for t in tasks),)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import multiprocessing
from typing import Optional

from flask import Response
from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer
from bentoml.saved_bundle import load_from_dir
from bentoml.server.api_server import BentoAPIServer
from bentoml.server.instruments import setup_prometheus_multiproc_dir

logger = logging.getLogger(__name__)


@inject
def gunicorn_bento_server(
    default_port: int = Provide[BentoMLContainer.config.bento_server.port],
    default_timeout: int = Provide[BentoMLContainer.config.bento_server.timeout],
    default_workers: int = Provide[BentoMLContainer.api_server_workers],
    default_enable_swagger: bool = Provide[
        BentoMLContainer.config.bento_server.swagger.enabled
    ],
    default_max_request_size: int = Provide[
        BentoMLContainer.config.bento_server.max_request_size
    ],
    default_loglevel=Provide[BentoMLContainer.config.bento_server.logging.level],
):
    from gunicorn.app.base import Application

    class GunicornBentoAPIServer(BentoAPIServer):
        def metrics_view_func(self):
            from prometheus_client import (
                CONTENT_TYPE_LATEST,
                CollectorRegistry,
                generate_latest,
                multiprocess,
            )

            registry = CollectorRegistry()
            # NOTE: enable mb metrics to be parsed.
            multiprocess.MultiProcessCollector(registry)
            return Response(generate_latest(registry), mimetype=CONTENT_TYPE_LATEST,)

    class GunicornBentoServer(Application):  # pylint: disable=abstract-method
        """
        A custom Gunicorn application.

        Usage::

            >>> from bentoml.server.gunicorn_server import GunicornBentoServer
            >>>
            >>> gunicorn_app = GunicornBentoServer(saved_bundle_path,\
            >>>     bind="127.0.0.1:5000")
            >>> gunicorn_app.run()

        :param bundle_path: path to the saved BentoService bundle
        :param bind: Specify a server socket to bind. Server sockets can be any of
            $(HOST), $(HOST):$(PORT), fd://$(FD), or unix:$(PATH).
        :param workers: number of worker processes
        :param timeout: request timeout config
        """

        @inject
        def __init__(
            self,
            bundle_path,
            bind: str = None,
            port: int = default_port,
            timeout: int = default_timeout,
            workers: int = default_workers,
            prometheus_lock: Optional[multiprocessing.Lock] = None,
            enable_swagger: bool = default_enable_swagger,
            max_request_size: int = default_max_request_size,
            loglevel: str = default_loglevel,
        ):
            self.bento_service_bundle_path = bundle_path

            if bind is None:
                self.bind = f"0.0.0.0:{port}"
            else:
                self.bind = bind

            self.options = {
                "bind": self.bind,
                "timeout": timeout,  # TODO
                "limit_request_line": max_request_size,
                "loglevel": loglevel.upper(),
            }
            if workers:
                self.options['workers'] = workers
            self.prometheus_lock = prometheus_lock
            self.enable_swagger = enable_swagger

            super(GunicornBentoServer, self).__init__()

        def load_config(self):
            self.load_config_from_file("python:bentoml.server.gunicorn_config")

            gunicorn_config = dict(
                [
                    (key, value)
                    for key, value in self.options.items()
                    if key in self.cfg.settings and value is not None
                ]
            )
            for key, value in gunicorn_config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            bento_service = load_from_dir(self.bento_service_bundle_path)
            api_server = GunicornBentoAPIServer(
                bento_service, enable_swagger=self.enable_swagger
            )
            return api_server.app

        def run(self):
            setup_prometheus_multiproc_dir(self.prometheus_lock)
            super(GunicornBentoServer, self).run()

    return GunicornBentoServer



# Copyright 2021 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from contextvars import ContextVar

span_context_var = ContextVar('span context', default=None)


def initialize_tracer(
    service_name, async_transport=False, host=None, port=0, sample_rate=1.0
):
    from opentracing.scope_managers.asyncio import (  # pylint: disable=E0401
        AsyncioScopeManager,
    )
    from jaeger_client.config import Config  # pylint: disable=E0401

    if sample_rate == 1.0:
        # sample all traces
        sampler_config = {'type': 'const', 'param': 1}
    elif sample_rate == 0.0:
        # sample none traces
        sampler_config = {'type': 'const', 'param': 0}
    else:
        # random sampling decision with the probability
        sampler_config = {'type': 'probabilistic', 'param': sample_rate}

    tracer_config = {
        'sampler': sampler_config,
    }

    if host:
        tracer_config['local_agent'] = {'reporting_host': host, 'reporting_port': port}

    config = Config(
        config=tracer_config,
        service_name=service_name,
        validate=True,
        scope_manager=AsyncioScopeManager() if async_transport else None,
    )

    return config.new_tracer()


class JaegerTracer:
    def __init__(self, address, port):
        self.address = address
        self.port = port

    @contextmanager
    def span(
        self,
        service_name,
        span_name,
        request_headers=None,
        async_transport=False,
        sample_rate=1.0,
        standalone=False,
        is_root=False,
    ):
        """
        Opentracing tracer function
        """
        from opentracing import Format  # pylint: disable=E0401
        from jaeger_client.constants import TRACE_ID_HEADER  # pylint: disable=E0401

        tracer = initialize_tracer(
            service_name, async_transport, self.address, self.port, sample_rate
        )
        if tracer is None:
            yield
            return

        span_context = None
        span_context_saved = span_context_var.get()
        if not is_root and not standalone:
            if request_headers is not None:
                span_context = tracer.extract(Format.HTTP_HEADERS, request_headers)

            if span_context is None:
                span_context = span_context_saved or None

        with tracer.start_active_span(
            operation_name=span_name, child_of=span_context
        ) as scope:
            if standalone:
                yield None
                return
            else:
                token = span_context_var.set(scope.span.context)
                if request_headers and TRACE_ID_HEADER not in request_headers:
                    tracer.inject(
                        scope.span.context, Format.HTTP_HEADERS, request_headers,
                    )
                yield scope
                span_context_var.reset(token)
                return

    @contextmanager
    def async_span(self, *args, **kwargs):
        with self.span(*args, async_transport=True, **kwargs) as ctx:
            yield ctx
        return


def get_jaeger_tracer(address, port):
    return JaegerTracer(address, port)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from bentoml.utils import status_pb_to_error_code_and_message
from bentoml.utils.lazy_loader import LazyLoader
from bentoml.cli.utils import Spinner
from bentoml.utils import get_default_yatai_client
from bentoml.cli.click_utils import (
    BentoMLCommandGroup,
    parse_bento_tag_callback,
    _echo,
    CLI_COLOR_SUCCESS,
)
from bentoml.cli.deployment import (
    _print_deployment_info,
    _print_deployments_info,
)
from bentoml.yatai.deployment import ALL_NAMESPACE_TAG
from bentoml.yatai.deployment.aws_ec2.constants import (
    DEFAULT_MIN_SIZE,
    DEFAULT_DESIRED_CAPACITY,
    DEFAULT_MAX_SIZE,
    DEFAULT_INSTANCE_TYPE,
    DEFAULT_AMI_ID,
)
from bentoml.exceptions import CLIException

yatai_proto = LazyLoader("yatai_proto", globals(), "bentoml.yatai.proto")


def get_aws_ec2_sub_command():
    # pylint: disable=unused-variable

    @click.group(name="ec2", cls=BentoMLCommandGroup, help="commands for EC2")
    def aws_ec2():
        pass

    @aws_ec2.command(help="Deploy BentoService to EC2")
    @click.argument("name", type=click.STRING)
    @click.option(
        "-b",
        "--bento",
        type=click.STRING,
        required=True,
        callback=parse_bento_tag_callback,
    )
    @click.option(
        "-n", "--namespace", type=click.STRING, callback=parse_bento_tag_callback,
    )
    @click.option(
        "--region", type=click.STRING, help="Region to deploy service in",
    )
    @click.option(
        "--min-size",
        type=click.INT,
        default=DEFAULT_MIN_SIZE,
        help="The minimum limit helps ensure that you always have a "
        "certain number of instances running at all times.Default is 1",
    )
    @click.option(
        "--desired-capacity",
        type=click.INT,
        default=DEFAULT_DESIRED_CAPACITY,
        help="Desired number of instances size to run BentoService on."
        "Should be between minimum and maximum capacities.Default is 1",
    )
    @click.option(
        "--max-size",
        type=click.INT,
        default=DEFAULT_MAX_SIZE,
        help="The maximum limit lets Amazon EC2 Auto Scaling scale out "
        "the number of instances as needed to handle an increase in demand. "
        "Default is 1",
    )
    @click.option(
        "--instance-type",
        type=click.STRING,
        default=DEFAULT_INSTANCE_TYPE,
        help="Instance type of EC2 container.Default is t2 micro",
    )
    @click.option(
        "--ami-id",
        type=click.STRING,
        default=DEFAULT_AMI_ID,
        help="AMI id.Default is Amazon Linux 2",
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will return immediately. The default value is wait',
    )
    def deploy(
        name,
        bento,
        namespace,
        region,
        min_size,
        desired_capacity,
        max_size,
        instance_type,
        ami_id,
        output,
        wait,
    ):
        _echo(
            message='AWS EC2 deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-ec2-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        bento_name, bento_version = bento.split(":")
        with Spinner(f"Deploying {bento} to AWS EC2"):
            result = yatai_client.deployment.create_ec2_deployment(
                name=name,
                namespace=namespace,
                bento_name=bento_name,
                bento_version=bento_version,
                region=region,
                min_size=min_size,
                desired_capacity=desired_capacity,
                max_size=max_size,
                instance_type=instance_type,
                ami_id=ami_id,
                wait=wait,
            )
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _print_deployment_info(result.deployment, output)
        _echo("Successfully created AWS EC2 deployment", CLI_COLOR_SUCCESS)

    @aws_ec2.command(help="Delete AWS EC2 deployment")
    @click.argument("name", type=click.STRING)
    @click.option(
        "-n",
        "--namespace",
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        "can be changed in BentoML configuration yatai_service/default_namespace",
    )
    @click.option(
        "--force",
        is_flag=True,
        help="force delete the deployment record in database and "
        "ignore errors when deleting cloud resources",
    )
    def delete(name, namespace, force):
        _echo(
            message='AWS EC2 deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-ec2-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        get_deployment_result = yatai_client.deployment.get(
            namespace=namespace, name=name
        )
        if get_deployment_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_deployment_result.status
            )
            raise CLIException(f"{error_code}:{error_message}")

        delete_deployment_result = yatai_client.deployment.delete(
            namespace=namespace, deployment_name=name, force_delete=force
        )
        if (
            delete_deployment_result.status.status_code
            != yatai_proto.status_pb2.Status.OK
        ):
            error_code, error_message = status_pb_to_error_code_and_message(
                delete_deployment_result.status
            )
            raise CLIException(f"{error_code}:{error_message}")

        _echo(f"Successfiully deleted AWS EC2 deployment '{name}'", CLI_COLOR_SUCCESS)

    @aws_ec2.command(help="Get EC2 deployment")
    @click.argument("name", type=click.STRING)
    @click.option(
        "-n",
        "--namespace",
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(  # pylint: disable=unused-variable
        "-o", "--output", type=click.Choice(["json", "yaml", "table"]), default="json"
    )
    def get(name, namespace, output):
        _echo(
            message='AWS EC2 deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-ec2-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        describe_result = yatai_client.deployment.describe(namespace, name)

        if describe_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise CLIException(f"{error_code}:{error_message}")

        get_result = yatai_client.deployment.get(namespace, name)
        if get_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise CLIException(f"{error_code}:{error_message}")

        _print_deployment_info(get_result.deployment, output)

    @aws_ec2.command(help="Update existing AWS EC2 deployments")
    @click.argument("name", type=click.STRING)
    @click.option(
        "-b",
        "--bento",
        "--bento-service-bundle",
        type=click.STRING,
        callback=parse_bento_tag_callback,
        help="Target BentoService to be deployed, referenced by its name and version "
        'in format of name:version. For example: "iris_classifier:v1.2.0"',
    )
    @click.option(
        "-n",
        "--namespace",
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        "--min-size",
        type=click.INT,
        default=DEFAULT_MIN_SIZE,
        help="The minimum limit helps ensure that you always have a "
        "certain number of instances running at all times.Default is 1",
    )
    @click.option(
        "--desired-capacity",
        type=click.INT,
        default=DEFAULT_DESIRED_CAPACITY,
        help="Desired number of instances to run BentoService on."
        "Should be between minimum and maximum capacities.Default is 1",
    )
    @click.option(
        "--max-size",
        type=click.INT,
        default=DEFAULT_MAX_SIZE,
        help="The maximum limit lets Amazon EC2 Auto Scaling scale out "
        "the number of instances as needed to handle an increase in demand. "
        "Default is 1",
    )
    @click.option(
        "--instance-type",
        type=click.STRING,
        default=DEFAULT_INSTANCE_TYPE,
        help="Instance type of EC2 container.Default is t2 micro",
    )
    @click.option(
        "--ami-id",
        type=click.STRING,
        default=DEFAULT_AMI_ID,
        help="AMI id.Default is Amazon Linux 2",
    )
    @click.option(
        "-o", "--output", type=click.Choice(["json", "yaml", "table"]), default="json"
    )
    @click.option(
        "--wait/--no-wait",
        default=True,
        help="Wait for apply action to complete or encounter an error."
        "If set to no-wait, CLI will return immediately. The default value is wait",
    )
    def update(
        name,
        bento,
        namespace,
        min_size,
        desired_capacity,
        max_size,
        instance_type,
        ami_id,
        output,
        wait,
    ):
        _echo(
            message='AWS EC2 deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-ec2-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        if bento:
            bento_name, bento_version = bento.split(":")
        else:
            bento_name = None
            bento_version = None

        with Spinner("Updating EC2 deployment"):
            update_result = yatai_client.deployment.update_ec2_deployment(
                deployment_name=name,
                bento_name=bento_name,
                bento_version=bento_version,
                namespace=namespace,
                min_size=min_size,
                desired_capacity=desired_capacity,
                max_size=max_size,
                instance_type=instance_type,
                ami_id=ami_id,
                wait=wait,
            )
            if update_result.status.status_code != yatai_proto.status_pb2.Status.OK:
                error_code, error_message = status_pb_to_error_code_and_message(
                    update_result.status
                )
                raise CLIException(f"{error_code}:{error_message}")

        _print_deployment_info(update_result.deployment, output)
        _echo(f"Successfiully updated AWS EC2 deployment '{name}'", CLI_COLOR_SUCCESS)

    @aws_ec2.command(name="list", help="List AWS Lambda deployments")
    @click.option(
        "-n",
        "--namespace",
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
        default=ALL_NAMESPACE_TAG,
    )
    @click.option(
        "--limit",
        type=click.INT,
        help="The maximum amount of AWS Lambda deployments to be listed at once",
    )
    @click.option(
        "--offset",
        type=click.INT,
        help="The offset for list of AWS Lambda deployments",
    )
    @click.option(
        "-l",
        "--labels",
        type=click.STRING,
        help="Label query to filter Lambda deployments, supports '=', '!=', 'IN', "
        "'NotIn', 'Exists', and 'DoesNotExist'. (e.g. key1=value1, "
        "key2!=value2, key3 In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        "--order-by", type=click.Choice(["created_at", "name"]), default="created_at",
    )
    @click.option(
        "--asc/--desc",
        default=False,  # pylint: disable=unused-variable
        help="Ascending or descending order for list deployments",
    )
    @click.option(
        "-o",
        "--output",
        type=click.Choice(["json", "yaml", "table", "wide"]),
        default="table",
    )
    def list_deployments(namespace, limit, offset, labels, order_by, asc, output):
        _echo(
            message='AWS EC2 deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-ec2-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        list_result = yatai_client.deployment.list_ec2_deployments(
            limit=limit,
            labels=labels,
            offset=offset,
            namespace=namespace,
            order_by=order_by,
            ascending_order=asc,
        )
        if list_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                list_result.status
            )
            raise CLIException(f"{error_code}:{error_message}")
        else:
            _print_deployments_info(list_result.deployments, output)

    return aws_ec2



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from dataclasses import asdict
from dataclasses import fields as get_fields
from dataclasses import is_dataclass


class DataclassJsonEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    def default(self, o):  # pylint: disable=method-hidden
        if is_dataclass(o):
            if hasattr(o, 'to_json'):
                return o.to_json()
            else:
                return asdict(o)
        return super().default(o)


class json_serializer:
    def __init__(self, fields=None, compat=False):
        self.fields = fields
        self.compat = compat

    @staticmethod
    def _extract_nested(obj):
        if hasattr(obj, "to_json"):
            return obj.to_json()
        return obj

    def __call__(self, klass):
        if not is_dataclass(klass):
            raise TypeError(
                f'{self.__class__.__name__} only accepts dataclasses, '
                f'got {klass.__name__}'
            )
        default_map = {
            f.name: f.default_factory() if callable(f.default_factory) else f.default
            for f in get_fields(klass)
        }
        if self.fields is None:
            self.fields = tuple(k for k in default_map.keys() if not k.startswith('_'))

        if self.compat:

            def to_json(data_obj):
                return {
                    k: self._extract_nested(getattr(data_obj, k))
                    for k in self.fields
                    if default_map[k] != getattr(data_obj, k)
                }

        else:

            def to_json(data_obj):
                return {
                    k: self._extract_nested(getattr(data_obj, k)) for k in self.fields
                }

        klass.to_json = to_json
        return klass



from contextlib import contextmanager
from bentoml.exceptions import LockUnavailable

import logging
import random
import time

from bentoml.yatai.db.stores.lock import LockStore, LOCK_STATUS

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT_SECONDS = 10
DEFAULT_TIMEOUT_JITTER_SECONDS = 1
DEFAULT_MAX_RETRY_COUNT = 5
DEFAULT_TTL_MIN = 3


# enum of read or write lock
class LockType:
    READ = LOCK_STATUS.read_lock
    WRITE = LOCK_STATUS.write_lock


@contextmanager
def lock(
    db,
    locks,
    timeout_seconds=DEFAULT_TIMEOUT_SECONDS,
    timeout_jitter_seconds=DEFAULT_TIMEOUT_JITTER_SECONDS,
    max_retry_count=DEFAULT_MAX_RETRY_COUNT,
    ttl_min=DEFAULT_TTL_MIN,
):
    """
    Context manager to acquire operation-level locks on all
    resources defined in the locks parameter
    :param db: instance of bentoml.yatai.db.DB
    :param locks: [
        (lock_identifier: string,
        lock_type: bentoml.yatai.locking.lock.LockType)
    ]
    :param timeout_seconds: amount of time to wait between lock acquisitions
    :param timeout_jitter_seconds: amount of random jitter to add to timeout
    :param max_retry_count: times to retry lock acquisition before failing
    :param ttl_min: amount of time before lock expires
    :return: (sess, lock_objs)
    :exception: LockUnavailable when lock cannot be acquired


    Example Usage:
    ```
    with lock(
        db, [(deployment_id, LockType.WRITE), (bento_id, LockType.READ)]
    ) as (sess, lock_objs):
        # begin critical section
    ```
    """
    if len(locks) < 1:
        raise ValueError("At least one lock needs to be acquired")

    # try to acquire lock
    for i in range(max_retry_count):
        try:
            # create session
            with db.create_session() as sess:
                lock_objs = []
                # acquire all locks in lock list
                for (lock_identifier, lock_type) in locks:
                    lock_obj = LockStore.acquire(
                        sess, lock_type, lock_identifier, ttl_min
                    )
                    lock_objs.append(lock_obj)

                # try to commit all locks to db
                sess.commit()
                logger.debug("Session acquired")
                for lck in locks:
                    op = 'READ' if lck[1] == LOCK_STATUS.read_lock else 'WRITE'
                    logger.debug(f"\t{op} on {lck[0]}")
                start = time.time()

                # return locked session to user
                yield sess, lock_objs

                # release all locks
                for lock_obj in lock_objs:
                    lock_obj.release(sess)
                sess.commit()

                end = time.time()
                logger.debug(f"Session released after {end - start}s")
                return
        except LockUnavailable as e:
            # wait before retrying
            sleep_seconds = timeout_seconds + random.random() * timeout_jitter_seconds
            time.sleep(sleep_seconds)
            logger.warning(
                f"Failed to acquire lock: {e}. "
                f"Retrying {max_retry_count - i - 1} more times..."
            )
    raise LockUnavailable(f"Failed to acquire lock after {max_retry_count} attempts")



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import Iterable, Iterator, Mapping, Optional, Sequence, Tuple

from bentoml.adapters.string_input import StringInput
from bentoml.exceptions import MissingDependencyException
from bentoml.types import HTTPHeaders, InferenceTask
from bentoml.utils.dataframe_util import (
    PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS,
    read_dataframes_from_json_n_csv,
)
from bentoml.utils.lazy_loader import LazyLoader

pandas = LazyLoader('pandas', globals(), 'pandas')

DataFrameTask = InferenceTask[str]
ApiFuncArgs = Tuple['pandas.DataFrame']


class DataframeInput(StringInput):
    """
    Convert various inputs(HTTP, Aws Lambda or CLI) to pandas dataframe, passing it to
    API functions.

    Parameters
    ----------
    orient : str
        Indication of expected JSON string format.
        Compatible JSON strings can be produced by ``to_json()`` with a
        corresponding orient value.
        The set of possible orients is:

        - ``'split'`` : dict like
          ``{index -> [index], columns -> [columns], data -> [values]}``
        - ``'records'`` : list like
          ``[{column -> value}, ... , {column -> value}]``
        - ``'index'`` : dict like ``{index -> {column -> value}}``
        - ``'columns'`` : dict like ``{column -> {index -> value}}``
        - ``'values'`` : just the values array

        The allowed and default values depend on the value
        of the `typ` parameter.

        * when ``typ == 'series'`` (not available now),

          - allowed orients are ``{'split','records','index'}``
          - default is ``'index'``
          - The Series index must be unique for orient ``'index'``.

        * when ``typ == 'frame'``,

          - allowed orients are ``{'split','records','index',
            'columns','values'}``
          - default is ``'columns'``
          - The DataFrame index must be unique for orients ``'index'`` and
            ``'columns'``.
          - The DataFrame columns must be unique for orients ``'index'``,
            ``'columns'``, and ``'records'``.

    typ : {'frame', 'series'}, default 'frame'
        ** Note: 'series' is not supported now. **
        The type of object to recover.

    dtype : dict, default None
        If is None, infer dtypes; if a dict of column to dtype, then use those.
        Not applicable for ``orient='table'``.

    input_dtypes : dict, default None
        ** Deprecated **
        The same as the `dtype`

    Raises
    -------
    ValueError: Incoming data is missing required columns in dtype

    ValueError: Incoming data format can not be handled. Only json and csv

    Examples
    -------
    Example Service:

    .. code-block:: python

        from bentoml import env, artifacts, api, BentoService
        from bentoml.adapters import DataframeInput
        from bentoml.frameworks.sklearn import SklearnModelArtifact

        @env(infer_pip_packages=True)
        @artifacts([SklearnModelArtifact('model')])
        class IrisClassifier(BentoService):

            @api(
                input=DataframeInput(
                    orient="records",
                    columns=["sw", "sl", "pw", "pl"],
                    dtype={"sw": "float", "sl": "float", "pw": "float", "pl": "float"},
                ),
                batch=True,
            )
            def predict(self, df):
                # Optional pre-processing, post-processing code goes here
                return self.artifacts.model.predict(df)

    Query with HTTP request::

        curl -i \\
          --header "Content-Type: application/json" \\
          --request POST \\
          --data '[{"sw": 1, "sl": 2, "pw": 1, "pl": 2}]' \\
          localhost:5000/predict

    OR::

        curl -i \\
          --header "Content-Type: text/csv" \\
          --request POST \\
          --data-binary @file.csv \\
          localhost:5000/predict

    Query with CLI command::

        bentoml run IrisClassifier:latest predict --input \\
          '[{"sw": 1, "sl": 2, "pw": 1, "pl": 2}]'

    OR::

        bentoml run IrisClassifier:latest predict --format csv --input-file test.csv

    """

    SINGLE_MODE_SUPPORTED = False

    def __init__(
        self,
        typ: str = "frame",
        orient: Optional[str] = None,
        columns: Sequence[str] = None,
        dtype: Mapping[str, object] = None,
        input_dtypes: Mapping[str, object] = None,
        **base_kwargs,
    ):
        super().__init__(**base_kwargs)
        dtype = dtype if dtype is not None else input_dtypes

        # Verify pandas imported properly and retry import if it has failed initially
        if pandas is None:
            raise MissingDependencyException(
                "Missing required dependency 'pandas' for DataframeInput, install "
                "with `pip install pandas`"
            )
        if typ != "frame":
            raise NotImplementedError()
        assert not orient or orient in PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS, (
            f"Invalid option 'orient'='{orient}', valid options are "
            f"{PANDAS_DATAFRAME_TO_JSON_ORIENT_OPTIONS}"
        )

        assert (
            columns is None or dtype is None or set(dtype) == set(columns)
        ), "dtype must match columns"

        self.typ = typ
        self.orient = orient
        self.columns = columns
        if isinstance(dtype, (list, tuple)):
            self.dtype = dict((index, dtype) for index, dtype in enumerate(dtype))
        else:
            self.dtype = dtype

    @property
    def pip_dependencies(self):
        return ['pandas']

    @property
    def config(self):
        base_config = super().config
        return dict(base_config, orient=self.orient, typ=self.typ, dtype=self.dtype,)

    def _get_type(self, item):
        if item.startswith('int'):
            return 'integer'
        if item.startswith('float') or item.startswith('double'):
            return 'number'
        if item.startswith('str') or item.startswith('date'):
            return 'string'
        if item.startswith('bool'):
            return 'boolean'
        return 'object'

    @property
    def request_schema(self):
        if isinstance(self.dtype, dict):
            json_schema = {  # For now, only declare JSON on docs.
                "schema": {
                    "type": "object",
                    "properties": {
                        k: {"type": "array", "items": {"type": self._get_type(v)}}
                        for k, v in self.dtype.items()
                    },
                }
            }
        else:
            json_schema = {"schema": {"type": "object"}}
        return {
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {"file": {"type": "string", "format": "binary"}},
                }
            },
            "application/json": json_schema,
            "text/csv": {"schema": {"type": "string", "format": "binary"}},
        }

    @classmethod
    def _detect_format(cls, task: InferenceTask) -> str:
        if task.aws_lambda_event:
            headers = HTTPHeaders.from_dict(task.aws_lambda_event.get('headers', {}))
            if headers.content_type == "application/json":
                return "json"
            if headers.content_type == "text/csv":
                return "csv"
        elif task.http_headers:
            headers = task.http_headers
            if headers.content_type == "application/json":
                return "json"
            if headers.content_type == "text/csv":
                return "csv"
        elif task.cli_args:
            parser = argparse.ArgumentParser()
            parser.add_argument('--format', type=str, choices=['csv', 'json'])
            parsed_args, _ = parser.parse_known_args(list(task.cli_args))
            return parsed_args.format or "json"

        return "json"

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask[str]]
    ) -> ApiFuncArgs:
        fmts, datas = tuple(
            zip(*((self._detect_format(task), task.data) for task in tasks))
        )

        df, batchs = read_dataframes_from_json_n_csv(
            datas, fmts, orient=self.orient, columns=self.columns, dtype=self.dtype,
        )

        if df is None:
            for task in tasks:
                task.discard(
                    http_status=400,
                    err_msg=f"{self.__class__.__name__} Wrong input format.",
                )
            return (df,)

        for task, batch, data in zip(tasks, batchs, datas):
            if batch == 0:
                task.discard(
                    http_status=400,
                    err_msg=f"{self.__class__.__name__} Wrong input format: {data}.",
                )
            else:
                task.batch = batch
        return (df,)

    def from_inference_job(
        self, input_=None, input_file=None, **extra_args,
    ) -> Iterator[InferenceTask[str]]:
        # TODO: generate small batches of InferenceTasks from large input files
        return super().from_inference_job(
            input_=input_, input_file=input_file, **extra_args
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class hybridmethod(object):
    """A decorator which allows definition of a Python object method with both
    instance-level and class-level behavior.

    """

    def __init__(self, instance_function):
        """Create a new :class:`.hybridmethod`.

       Usage is typically via decorator::

           from bentoml.util.hybridmethod import hybridmethod

           class SomeClass(object):
               @hybridmethod
               def func(self, x, y):
                   return self._value + x + y

               @func.classmethod
               def value(cls, x, y):
                   return cls._default_value + x + y
       """
        self._instance_function = instance_function
        self._class_function = None

    def __get__(self, instance, owner):
        if instance is None:
            assert self._class_function is not None, "classmethod not defined"
            return self._class_function.__get__(owner, type(owner))
        else:
            return self._instance_function.__get__(instance, owner)

    def classmethod(self, class_function):
        self._class_function = class_function
        if not self._class_function.__doc__:
            self._class_function.__doc__ = self._instance_function.__doc__
        return self

    def __call__(self, *args, **kwargs):
        """Needed for this to work in python2.7
        """
        return self.__get__(self, type(self))(*args, **kwargs)



import os
import re
import logging
from typing import List
from pathlib import Path

from bentoml.exceptions import InvalidArgument, FailedPrecondition
from bentoml.service.env import BentoServiceEnv
from bentoml.utils.ruamel_yaml import YAML

logger = logging.getLogger(__name__)

ARTIFACTS_DIR_NAME = "artifacts"


class BentoServiceArtifact:
    """
    BentoServiceArtifact is the base abstraction for describing the trained model
    serialization and deserialization process.
    """

    def __new__(cls, *args, **kwargs):
        obj = super(BentoServiceArtifact, cls).__new__(cls)

        # store the args and kwargs used for creating this artifact instance - this is
        # used internally for the _copy method below

        # TODO: this warning may due to pylint's bug (issue 4668). We may
        # remove the following comment later
        # pylint: disable=unused-private-member
        obj.__args = args
        obj.__kwargs = kwargs

        return obj

    def _copy(self):
        """
        Create a new empty artifact instance with the same name and arguments, this
        is only used internally for BentoML to create new artifact instances when
        creating a new BentoService instance. BentoML stores a reference to the
        arguments provided for initializing this artifact class for the purpose of
        recreating a new instance with the same config
        """
        return self.__class__(*self.__args, **self.__kwargs)

    def __init__(self, name: str):
        if not name.isidentifier():
            raise ValueError(
                "Artifact name must be a valid python identifier, a string \
                is considered a valid identifier if it only contains \
                alphanumeric letters (a-z) and (0-9), or underscores (_). \
                A valid identifier cannot start with a number, or contain any spaces."
            )
        self._name = name
        self._packed = False
        self._loaded = False
        self._metadata = dict()

    @property
    def packed(self):
        return self._packed

    @property
    def loaded(self):
        return self._loaded

    @property
    def metadata(self):
        return self._metadata

    @property
    def is_ready(self):
        return self.packed or self.loaded

    @property
    def name(self):
        """
        The name of an artifact. It can be used to reference this artifact in a
        BentoService inference API callback function, via `self.artifacts[NAME]`
        """
        return self._name

    def pack(self, model, metadata: dict = None):  # pylint: disable=unused-argument
        """
        Pack the in-memory trained model object to this BentoServiceArtifact

        Note: add "# pylint:disable=arguments-differ" to child class's pack method
        """

    def load(self, path):
        """
        Load artifact assuming it was 'self.save' on the same `path`
        """

    def _metadata_path(self, base_path):
        return os.path.join(
            base_path, re.sub("[^-a-zA-Z0-9_.() ]+", "", self.name) + ".yml",
        )

    def save(self, dst):
        """
        Save artifact to given dst path

        In a BentoService saved bundle, all artifacts are being saved to the
        `{service.name}/artifact/` directory. Thus, `save` must use the artifact name
        as a way to differentiate its files from other artifacts.

        For example, a BentoServiceArtifact class' save implementation should not create
        a file named `config.json` in the `dst` directory. Because when a BentoService
        has multiple artifacts of this same artifact type, they will all try to create
        this `config.json` file and overwrite each other. The right way to do it here,
        is use the name as the prefix(or part of the file name), e.g.
         f"{artifact.name}-config.json"
        """

    def get(self):
        """
        Get returns a reference to the artifact being packed or loaded from path
        """

    def set_dependencies(self, env: BentoServiceEnv):
        """modify target BentoServiceEnv instance to ensure the required dependencies
        are listed in the BentoService environment spec

        :param env: target BentoServiceEnv instance to modify
        """

    def __getattribute__(self, item):
        if item == 'pack':
            original = object.__getattribute__(self, item)

            def wrapped_pack(*args, **kwargs):
                if self.packed:
                    logger.warning(
                        "`pack` an artifact multiple times may lead to unexpected "
                        "behaviors"
                    )
                if 'metadata' in kwargs:
                    if isinstance(kwargs['metadata'], dict):
                        self._metadata = kwargs['metadata']
                    else:
                        raise TypeError(
                            "Setting a non-dictionary metadata " "is not supported."
                        )
                ret = original(*args, **kwargs)
                # do not set `self._pack` if `pack` has failed with an exception raised
                self._packed = True
                return ret

            return wrapped_pack

        elif item == 'load':
            original = object.__getattribute__(self, item)

            def wrapped_load(*args, **kwargs):
                if self.packed:
                    logger.warning(
                        "`load` on a 'packed' artifact may lead to unexpected behaviors"
                    )
                if self.loaded:
                    logger.warning(
                        "`load` an artifact multiple times may lead to unexpected "
                        "behaviors"
                    )

                # load metadata if exists
                path = args[0]  # load(self, path)
                meta_path = self._metadata_path(path)
                if os.path.isfile(meta_path):
                    with open(meta_path) as file:
                        yaml = YAML()
                        yaml_content = file.read()
                        self._metadata = yaml.load(yaml_content)

                ret = original(*args, **kwargs)
                # do not set self._loaded if `load` has failed with an exception raised
                self._loaded = True
                return ret

            return wrapped_load

        elif item == 'save':

            def wrapped_save(*args, **kwargs):
                if not self.is_ready:
                    raise FailedPrecondition(
                        "Trying to save empty artifact. An artifact needs to be `pack` "
                        "with model instance or `load` from saved path before saving"
                    )

                # save metadata
                dst = args[0]  # save(self, dst)
                if self.metadata:
                    yaml = YAML()
                    yaml.dump(self.metadata, Path(self._metadata_path(dst)))

                original = object.__getattribute__(self, item)
                return original(*args, **kwargs)

            return wrapped_save

        elif item == 'get':

            def wrapped_get(*args, **kwargs):
                if not self.is_ready:
                    raise FailedPrecondition(
                        "Trying to access empty artifact. An artifact needs to be "
                        "`pack` with model instance or `load` from saved path before "
                        "it can be used for inference"
                    )
                original = object.__getattribute__(self, item)
                return original(*args, **kwargs)

            return wrapped_get

        return object.__getattribute__(self, item)


class ArtifactCollection(dict):
    """
    A dict of artifact instances (artifact.name -> artifact_instance)
    """

    def __setitem__(self, key: str, artifact: BentoServiceArtifact):
        if key != artifact.name:
            raise InvalidArgument(
                "Must use Artifact name as key, {} not equal to {}".format(
                    key, artifact.name
                )
            )
        self.add(artifact)

    def __getattr__(self, item: str):
        """Proxy to `BentoServiceArtifact#get()` to allow easy access to the model
        artifact in its native form. E.g. for a BentoService class with an artifact
        `SklearnModelArtifact('model')`, when user code accesses `self.artifacts.model`
        in the BentoService API callback function, instead of returning an instance of
        BentoServiceArtifact, it returns a Sklearn model object
        """
        return self[item].get()

    def get(self, item: str) -> BentoServiceArtifact:
        """Access the BentoServiceArtifact instance by artifact name
        """
        return self[item]

    def get_artifact_list(self) -> List[BentoServiceArtifact]:
        """Access the list of BentoServiceArtifact instances
        """
        return super(ArtifactCollection, self).values()

    def add(self, artifact: BentoServiceArtifact):
        super(ArtifactCollection, self).__setitem__(artifact.name, artifact)

    def save(self, dst):
        """Save all BentoServiceArtifact instances in self.values() to `dst` path
        if they are `packed` or `loaded`
        """
        save_path = os.path.join(dst, ARTIFACTS_DIR_NAME)
        os.mkdir(save_path)
        for artifact in self.get_artifact_list():
            if artifact.is_ready:
                artifact.save(save_path)
            else:
                logger.warning(
                    "Skip saving empty artifact, %s('%s')",
                    artifact.__class__.__name__,
                    artifact.name,
                )

    @classmethod
    def from_declared_artifact_list(cls, artifacts_list: List[BentoServiceArtifact]):
        artifact_collection = cls()
        for artifact in artifacts_list:
            artifact_collection.add(artifact._copy())
        return artifact_collection

    def load_all(self, path):
        for artifact in self.values():
            artifact.load(path)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

from bentoml.utils.ruamel_yaml import YAML
from bentoml.yatai.proto.deployment_pb2 import Deployment, DeploymentSpec
from bentoml.exceptions import InvalidArgument, YataiDeploymentException

logger = logging.getLogger(__name__)

SPEC_FIELDS_AVAILABLE_FOR_UPDATE = ['bento_name', 'bento_version']

SAGEMAKER_FIELDS_AVAILABLE_FOR_UPDATE = [
    'api_name',
    'instance_type',
    'instance_count',
    'num_of_gunicorn_workers_per_instance',
]


def deployment_dict_to_pb(deployment_dict):
    deployment_pb = Deployment()
    if deployment_dict.get('spec'):
        spec_dict = deployment_dict.get('spec')
    else:
        raise YataiDeploymentException('"spec" is required field for deployment')
    platform = spec_dict.get('operator')
    if platform is not None:
        # converting platform parameter to DeploymentOperator name in proto
        # e.g. 'aws-lambda' to 'AWS_LAMBDA'
        deployment_pb.spec.operator = DeploymentSpec.DeploymentOperator.Value(
            platform.replace('-', '_').upper()
        )

    for field in ['name', 'namespace']:
        if deployment_dict.get(field):
            deployment_pb.__setattr__(field, deployment_dict.get(field))
    if deployment_dict.get('labels') is not None:
        deployment_pb.labels.update(deployment_dict.get('labels'))
    if deployment_dict.get('annotations') is not None:
        deployment_pb.annotations.update(deployment_dict.get('annotations'))

    if spec_dict.get('bento_name'):
        deployment_pb.spec.bento_name = spec_dict.get('bento_name')
    if spec_dict.get('bento_version'):
        deployment_pb.spec.bento_version = spec_dict.get('bento_version')

    if deployment_pb.spec.operator == DeploymentSpec.AWS_SAGEMAKER:
        sagemaker_config = spec_dict.get('sagemaker_operator_config', {})
        sagemaker_config_pb = deployment_pb.spec.sagemaker_operator_config
        for field in [
            'region',
            'api_name',
            'instance_type',
            'num_of_gunicorn_workers_per_instance',
            'instance_count',
        ]:
            if sagemaker_config.get(field):
                sagemaker_config_pb.__setattr__(field, sagemaker_config.get(field))
    elif deployment_pb.spec.operator == DeploymentSpec.AWS_LAMBDA:
        lambda_conf = spec_dict.get('aws_lambda_operator_config', {})
        for field in ['region', 'api_name', 'memory_size', 'timeout']:
            if lambda_conf.get(field):
                deployment_pb.spec.aws_lambda_operator_config.__setattr__(
                    field, lambda_conf.get(field)
                )
    elif deployment_pb.spec.operator == DeploymentSpec.AZURE_FUNCTIONS:
        azure_functions_config = spec_dict.get('azure_function_operators_config', {})
        for field in [
            'location',
            'min_instances',
            'max_burst',
            'premium_plan_sku',
            'function_auth_level',
        ]:
            if azure_functions_config.get(field):
                deployment_pb.spec.azure_functions_operator_config.__setattr__(
                    field, azure_functions_config.get(field)
                )
    else:
        raise InvalidArgument(
            'Platform "{}" is not supported in the current version of '
            'BentoML'.format(platform)
        )

    return deployment_pb


def deployment_yaml_string_to_pb(deployment_yaml_string):
    yaml = YAML()
    deployment_yaml = yaml.load(deployment_yaml_string)
    return deployment_dict_to_pb(deployment_yaml)



import os
from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


class DetectronModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading Detectron2 models,
    in the form of detectron2.checkpoint.DetectionCheckpointer

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: detectron2 package is required for
        DetectronModelArtifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            torch.nn.Module

    Example usage:

    >>> # Train model with data
    >>>
    >>>
    >>> import bentoml
    >>> from bentoml.adapters import ImageInput
    >>> from bentoml.frameworks.detectron import DetectronModelArtifact
    >>> from detectron2.data import transforms as T
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([DetectronModelArtifact('model')])
    >>> class CocoDetectronService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=ImageInput(), batch=False)
    >>>     def predict(self, img: np.ndarray) -> Dict:
    >>>         _aug = T.ResizeShortestEdge(
    >>>            [800, 800], 1333
    >>>        )
    >>>        boxes = None
    >>>        scores = None
    >>>        pred_classes = None
    >>>        pred_masks= None
    >>>        try:
    >>>            original_image = img[:, :, ::-1]
    >>>            height, width = original_image.shape[:2]
    >>>            image = _aug.get_transform(original_image).
    >>>                             apply_image(original_image)
    >>>            image = torch.as_tensor(image.astype("float32").
    >>>                             transpose(2, 0, 1))
    >>>            inputs = {"image": image, "height": height, "width": width}
    >>>            predictions = self.artifacts.model([inputs])[0]
    >>>
    >>>
    >>> cfg = get_cfg()
    >>> cfg.merge_from_file("input_model.yaml")
    >>> meta_arch = META_ARCH_REGISTRY.get(cfg.MODEL.META_ARCHITECTURE)
    >>> model = meta_arch(cfg)
    >>> model.eval()
    >>> device = "cuda:{}".format(0)
    >>> model.to(device)
    >>> checkpointer = DetectionCheckpointer(model)
    >>> checkpointer.load("output/model.pth")
    >>> metadata = {
    >>>     'device' : device
    >>> }
    >>> bento_svc = CocoDetectronService()
    >>> bento_svc.pack('model', model, metadata, "input_model.yaml")
    >>> saved_path = bento_svc.save()
    """

    def __init__(self, name):
        super().__init__(name)
        self._model = None
        self._aug = None
        self._input_model_yaml = None

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['torch', "detectron2"])

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def pack(
        self, model, metadata=None, input_model_yaml=None
    ):  # pylint:disable=arguments-differ
        try:
            import detectron2  # noqa # pylint: disable=unused-import
        except ImportError:
            raise MissingDependencyException(
                "Detectron package is required to use DetectronModelArtifact"
            )
        self._model = model
        self._metadata = metadata
        self._input_model_yaml = input_model_yaml
        return self

    def load(self, path):
        try:
            from detectron2.checkpoint import (
                DetectionCheckpointer,
            )  # noqa # pylint: disable=unused-import
            from detectron2.modeling import META_ARCH_REGISTRY
            from detectron2.config import get_cfg
            from detectron2.data import transforms as T
        except ImportError:
            raise MissingDependencyException(
                "Detectron package is required to use DetectronArtifact"
            )
        cfg = get_cfg()
        cfg.merge_from_file(f"{path}/{self.name}.yaml")
        meta_arch = META_ARCH_REGISTRY.get(cfg.MODEL.META_ARCHITECTURE)
        self._model = meta_arch(cfg)
        self._model.eval()

        device = self._metadata['device']
        self._model.to(device)
        checkpointer = DetectionCheckpointer(self._model)
        checkpointer.load(f"{path}/{self.name}.pth")
        self._aug = T.ResizeShortestEdge(
            [cfg.INPUT.MIN_SIZE_TEST, cfg.INPUT.MIN_SIZE_TEST], cfg.INPUT.MAX_SIZE_TEST
        )
        return self.pack(self._model)

    def get(self):
        return self._model

    def save(self, dst):
        try:
            from detectron2.checkpoint import (
                DetectionCheckpointer,
            )  # noqa # pylint: disable=unused-import
            from detectron2.config import get_cfg
        except ImportError:
            raise MissingDependencyException(
                "Detectron package is required to use DetectronArtifact"
            )
        os.makedirs(dst, exist_ok=True)
        checkpointer = DetectionCheckpointer(self._model, save_dir=dst)
        checkpointer.save(self.name)
        cfg = get_cfg()
        cfg.merge_from_file(self._input_model_yaml)
        with open(
            os.path.join(dst, f"{self.name}.yaml"), 'w', encoding='utf-8'
        ) as output_file:
            output_file.write(cfg.dump())



import os
import shutil

from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


class OnnxMlirModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading onnx-mlir compiled model and operationalized
    using pyruntime wrapper

    onnx-mlir is a compiler technology that can take an onnx model and lower it
    (using llvm) to an inference library that is optimized and has little external
    dependencies.

    The PyRuntime interface is created during the build of onnx-mlir using pybind.
    See the onnx-mlir supporting documentation for detail.

    Args:
        name (string): Name of the artifact
    Raises:
        MissingDependencyException: PyRuntime must be accessible in path.

    Example usage:

    >>> import bentoml
    >>> from bentoml import env, artifacts, api, BentoService
    >>> from bentoml.adapters import ImageInput
    >>> from bentoml.frameworks.onnxmlir import OnnxMlirModelArtifact
    >>> from bentoml.types import JsonSerializable, InferenceTask, InferenceError
    >>> from bentoml.service.artifacts.common import PickleArtifact
    >>> from typing import List
    >>> import numpy as np
    >>> import pandas as pd
    >>> import sys
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([OnnxMlirModelArtifact('model'), PickleArtifact('labels')])
    >>> class ResNetPredict(BentoService):
    >>>
    >>>     def preprocess(self, input_data):
    >>>         # convert the input data into the float32 input
    >>>         img_data = np.stack(input_data).transpose(0, 3, 1, 2)
    >>>
    >>>         #normalize
    >>>         mean_vec = np.array([0.485, 0.456, 0.406])
    >>>         stddev_vec = np.array([0.229, 0.224, 0.225])
    >>>
    >>>
    >>>         norm_img_data = np.zeros(img_data.shape).astype('float32')
    >>>
    >>>
    >>>         for i in range(img_data.shape[0]):
    >>>             for j in range(img_data.shape[1]):
    >>>                 norm_img_data[i,j,:,:] =
    >>>                 (img_data[i,j,:,:]/255 - mean_vec[j]) / stddev_vec[j]
    >>>
    >>>         #add batch channel
    >>>         norm_img_data = norm_img_data.reshape(-1, 3, 224, 224).astype('float32')
    >>>         return norm_img_data
    >>>
    >>>
    >>>     def softmax(self, x):
    >>>         x = x.reshape(-1)
    >>>         e_x = np.exp(x - np.max(x))
    >>>         return e_x / e_x.sum(axis=0)
    >>>
    >>>     def post_process(self, raw_result):
    >>>         return self.softmax(np.array(raw_result)).tolist()
    >>>
    >>>     @bentoml.api(input=ImageInput(), batch=True)
    >>>     def predict(self, image_ndarrays: List[np.ndarray]) -> List[str]:
    >>>         input_datas = self.preprocess(image_ndarrays)
    >>>
    >>>         outputs = []
    >>>         for i in range(input_datas.shape[0]):
    >>>             raw_result = self.artifacts.model.run(input_datas[i:i+1])
    >>>             result = self.post_process(raw_result)
    >>>             idx = np.argmax(result)
    >>>             sort_idx = np.flip(np.squeeze(np.argsort(result)))
    >>>
    >>>             # return top 5 labels
    >>>             outputs.append(self.artifacts.labels[sort_idx[:5]])
    >>>             return outputs
    >>>
    """

    def __init__(self, name):
        super().__init__(name)
        self._inference_session = None
        self._model_so_path = None

    def _saved_model_file_path(self, base_path):
        self._model_so_path = os.path.join(base_path, self.name + '.so')
        return os.path.join(base_path, self.name + '.so')

    def pack(self, onnxmlir_model_so, metadata=None):
        # pylint:disable=arguments-renamed
        self._model_so_path = onnxmlir_model_so
        return self

    def load(self, path):
        print(path)
        return self.pack(self._saved_model_file_path(path))

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(["numpy"])

    def _get_onnxmlir_inference_session(self):
        try:
            # this has to be able to find the arch and OS specific PyRuntime .so file
            from PyRuntime import ExecutionSession
        except ImportError:
            raise MissingDependencyException(
                "PyRuntime package library must be in python path"
            )
        return ExecutionSession(self._model_so_path, "run_main_graph")

    def get(self):
        if not self._inference_session:
            self._inference_session = self._get_onnxmlir_inference_session()
        return self._inference_session

    def save(self, dst):
        # copies the model .so and places in the version controlled deployment path
        shutil.copyfile(self._model_so_path, self._saved_model_file_path(dst))



import logging
import os
import zipfile
import pathlib
import shutil

from bentoml.exceptions import (
    InvalidArgument,
    MissingDependencyException,
)
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv
from bentoml.utils import cloudpickle

logger = logging.getLogger(__name__)


def _is_path_like(path):
    return isinstance(path, (str, bytes, pathlib.Path, os.PathLike))


def _is_pytorch_lightning_model_file_like(path):
    return (
        _is_path_like(path)
        and os.path.isfile(path)
        and str(path).lower().endswith(".pt")
    )


class PytorchModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading objects with torch.save and torch.load

    Args:
        name (string): name of the artifact

    Raises:
        MissingDependencyException: torch package is required for PytorchModelArtifact
        InvalidArgument: invalid argument type, model being packed must be instance of
            torch.nn.Module

    Example usage:

    >>> import torch.nn as nn
    >>>
    >>> class Net(nn.Module):
    >>>     def __init__(self):
    >>>         super(Net, self).__init__()
    >>>         ...
    >>>
    >>>     def forward(self, x):
    >>>         ...
    >>>
    >>> net = Net()
    >>> # Train model with data
    >>>
    >>>
    >>> import bentoml
    >>> from bentoml.adapters import ImageInput
    >>> from bentoml.frameworks.pytorch import PytorchModelArtifact
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([PytorchModelArtifact('net')])
    >>> class PytorchModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=ImageInput(), batch=True)
    >>>     def predict(self, imgs):
    >>>         outputs = self.artifacts.net(imgs)
    >>>         return outputs
    >>>
    >>>
    >>> svc = PytorchModelService()
    >>>
    >>> # Pytorch model can be packed directly.
    >>> svc.pack('net', net)
    >>>
    >>> # Alternatively,
    >>>
    >>> # Pack a TorchScript Model
    >>> # Random input in the format expected by the net
    >>> sample_input = ...
    >>> traced_net = torch.jit.trace(net, sample_input)
    >>> svc.pack('net', traced_net)
    """

    def __init__(self, name, file_extension=".pt"):
        super().__init__(name)
        self._file_extension = file_extension
        self._model = None

    def _file_path(self, base_path):
        return os.path.join(base_path, self.name + self._file_extension)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-renamed
        try:
            import torch
        except ImportError:
            raise MissingDependencyException(
                "torch package is required to use PytorchModelArtifact"
            )

        if not isinstance(model, torch.nn.Module):
            raise InvalidArgument(
                "PytorchModelArtifact can only pack type \
                'torch.nn.Module' or 'torch.jit.ScriptModule'"
            )

        self._model = model
        return self

    def load(self, path):
        try:
            import torch
        except ImportError:
            raise MissingDependencyException(
                "torch package is required to use PytorchModelArtifact"
            )

        # TorchScript Models are saved as zip files
        if zipfile.is_zipfile(self._file_path(path)):
            model = torch.jit.load(self._file_path(path))
        else:
            model = cloudpickle.load(open(self._file_path(path), 'rb'))

        if not isinstance(model, torch.nn.Module):
            raise InvalidArgument(
                "Expecting PytorchModelArtifact loaded object type to be "
                "'torch.nn.Module' or 'torch.jit.ScriptModule' \
                but actually it is {}".format(
                    type(model)
                )
            )

        return self.pack(model)

    def set_dependencies(self, env: BentoServiceEnv):
        logger.warning(
            "BentoML by default does not include spacy and torchvision package when "
            "using PytorchModelArtifact. To make sure BentoML bundle those packages if "
            "they are required for your model, either import those packages in "
            "BentoService definition file or manually add them via "
            "`@env(pip_packages=['torchvision'])` when defining a BentoService"
        )
        if env._infer_pip_packages:
            env.add_pip_packages(['torch'])

    def get(self):
        return self._model

    def save(self, dst):
        try:
            import torch
        except ImportError:
            raise MissingDependencyException(
                "torch package is required to use PytorchModelArtifact"
            )

        # If model is a TorchScriptModule, we cannot apply standard pickling
        if isinstance(self._model, torch.jit.ScriptModule):
            return torch.jit.save(self._model, self._file_path(dst))

        return cloudpickle.dump(self._model, open(self._file_path(dst), "wb"))


class PytorchLightningModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading pytorch lightning model

    Args:
        name (string): Name of the pytorch model
    Raises:
        MissingDependencyException: torch and pytorch_lightning package is required.

    Example usage:

    >>>
    >>> # Train pytorch lightning model
    >>> from pytorch_lightning.core.lightning import LightningModule
    >>>
    >>> class SimpleModel(LightningModule):
    >>>     def forward(self, x):
    >>>         return x.add(1)
    >>>
    >>> model = SimpleModel()
    >>>
    >>> import bentoml
    >>> import torch
    >>> from bentoml.adapters import JsonInput
    >>> from bentoml.frameworks.pytorch import PytorchLightningModelArtifact
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([PytorchLightningModelArtifact('model')])
    >>> class PytorchLightingService(bentoml.BentoService):
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         input_tensor = torch.from_numpy(df.to_numpy())
    >>>         return self.artifacts.model(input).numpy()
    >>>
    >>> svc = PytorchLightingService()
    >>> # Pack Pytorch Lightning model instance
    >>> svc.pack('model', model)
    >>>
    >>> # Pack saved Pytorch Lightning model
    >>> # import torch
    >>> # script = model.to_torchscript()
    >>> # saved_model_path = 'path/to/model.pt'
    >>> # torch.jit.save(script, saved_model_path)
    >>> # svc.pack('model', saved_model_path)
    >>>
    >>> svc.save()
    """

    def __init__(self, name):
        super().__init__(name)
        self._model = None
        self._model_path = None

    def _saved_model_file_path(self, base_path):
        return os.path.join(base_path, self.name + '.pt')

    def pack(self, path_or_model, metadata=None):  # pylint:disable=arguments-renamed
        if _is_pytorch_lightning_model_file_like(path_or_model):
            logger.info(
                'PytorchLightningArtifact is packing a saved torchscript module '
                'from path'
            )
            self._model_path = path_or_model
        else:
            try:
                from pytorch_lightning.core.lightning import LightningModule
            except ImportError:
                raise InvalidArgument(
                    '"pytorch_lightning.lightning.LightningModule" model is required '
                    'to pack a PytorchLightningModelArtifact'
                )
            if isinstance(path_or_model, LightningModule):
                logger.info(
                    'PytorchLightningArtifact is packing a pytorch lightning '
                    'model instance as torchscript module'
                )
                self._model = path_or_model.to_torchscript()
            else:
                raise InvalidArgument(
                    'a LightningModule model is required to pack a '
                    'PytorchLightningModelArtifact'
                )
        return self

    def load(self, path):
        self._model = self._get_torch_script_model(self._saved_model_file_path(path))

    def set_dependencies(self, env: BentoServiceEnv):
        env.add_pip_packages(['pytorch-lightning'])

    def get(self):
        if self._model is None:
            self._model = self._get_torch_script_model(self._model_path)
        return self._model

    def save(self, dst):
        if self._model:
            try:
                import torch
            except ImportError:
                raise MissingDependencyException(
                    '"torch" package is required for saving Pytorch lightning model'
                )
            torch.jit.save(self._model, self._saved_model_file_path(dst))
        if self._model_path:
            shutil.copyfile(self._model_path, self._saved_model_file_path(dst))

    @staticmethod
    def _get_torch_script_model(model_path):
        try:
            from torch import jit
        except ImportError:
            raise MissingDependencyException(
                '"torch" package is required for inference with '
                'PytorchLightningModelArtifact'
            )
        return jit.load(model_path)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import traceback
from typing import BinaryIO, Iterable, Sequence, Tuple

from bentoml.adapters.multi_file_input import MultiFileInput
from bentoml.adapters.utils import (
    check_file_extension,
    get_default_accept_image_formats,
)
from bentoml.types import InferenceTask
from bentoml.utils.lazy_loader import LazyLoader

# BentoML optional dependencies, using lazy load to avoid ImportError
imageio = LazyLoader('imageio', globals(), 'imageio')
numpy = LazyLoader('numpy', globals(), 'numpy')


MultiImgTask = InferenceTask[Tuple[BinaryIO, ...]]  # image file bytes, json bytes
ApiFuncArgs = Tuple[Sequence['numpy.ndarray'], ...]


class MultiImageInput(MultiFileInput):
    """
    Transform incoming image data from http request, cli or lambda event into imageio
    arrays (a subclass of numpy.ndarray that has a meta attribute) and pass down to
    user defined API functions.

    Parameters
    ----------
    input_names : List[str]
        A tuple of acceptable input name for HTTP request.
        Default value is (image,)

    accept_image_formats : List[str]
        A list of acceptable image formats.
        Default value is loaded from bentoml config
        'apiserver/default_image_input_accept_file_extensions', which is
        set to ['.jpg', '.png', '.jpeg', '.tiff', '.webp', '.bmp'] by default.
        List of all supported format can be found here:
        https://imageio.readthedocs.io/en/stable/formats.html

    pilmode : str
        The pilmode to be used for reading image file into numpy
        array. Default value is 'RGB'.  Find more information at:
        https://imageio.readthedocs.io/en/stable/format_png-pil.html

    Raises
    ----------
    ImportError: imageio package is required to use MultiImageInput

    Examples
    ----------

    Service using MultiImageInput:

    .. code-block:: python

        from bentoml import BentoService
        import bentoml

        class MyService(BentoService):
            @bentoml.api(
                input=MultiImageInput(input_names=('imageX', 'imageY')), batch=True)
            def predict(self, image_groups):
                for image_group in image_groups:
                    image_array_x = image_group['imageX']
                    image_array_y = image_group['imageY']

    Query the endpoint with HTTP request performed by cURL::

        curl -i \\
          -F imageX=@testx.jpg \\
          -F imageY=@testy.jpg \\
          localhost:5000/predict

    OR by an HTML form that sends multipart data:

    .. code-block:: html

        <form action="http://localhost:8000" method="POST"
              enctype="multipart/form-data">
            <input name="imageX" type="file">
            <input name="imageY" type="file">
            <input type="submit">
        </form>

    OR by Python requests:

    .. code-block:: python

        import requests

        with open("testx.jpg", "rb") as f:
            image_bytes1 = f.read()
        with open("testy.jpg", "rb") as f:
            image_bytes2 = f.read()

        files = {
            "imageX": ("testx.jpg", image_bytes1),
            "imageY": ("testy.jpg", image_bytes2),
        }
        response = requests.post(your_url, files=files)

    .. code-block:: python

        import requests
        import PIL

        pil_image_x = PIL.Image.open('testx.jpg')
        pil_image_y = PIL.Image.open('testy.jpg')

        image_bytes1 = pil_image_x.tobytes()  # from PIL.Image
        image_bytes1 = pil_image_y.tobytes()

        files = {
            "imageX": ("testx.jpg", image_bytes1),
            "imageY": ("testy.jpg", image_bytes2),
        }
        response = requests.post(your_url, files=files)

    Query with CLI command::

        bentoml run PyTorchFashionClassifier:latest predict \\
          --input-file-imageX testx.jpg \\
          --input-file-imageY testy.jpg

    OR infer all file pairs under a folder with ten pairs each batch::

        bentoml run PyTorchFashionClassifier:latest predict --max-batch-size 10 \\
          --input-file-imageX folderx/*.jpg \\
          --input-file-imageY foldery/*.jpg

    Note: jpg files and json files should be in same prefix like this::

        folderx:
            - 1.jpg
            - 2.jpg
            ...
        foldery:
            - 1.jpg
            - 2.jpg
            ...


    Or the following cURL command

    >>> curl -F imageX=@image_file_x.png
    >>>      -F imageY=@image_file_y.jpg
    >>>      http://localhost:8000
    """

    def __init__(
        self,
        input_names=("image",),
        pilmode="RGB",
        accept_image_formats=None,
        **base_kwargs,
    ):
        assert imageio, "`imageio` dependency can be imported"

        super().__init__(input_names=input_names, allow_none=True, **base_kwargs)

        self.pilmode = pilmode
        self.accept_image_formats = (
            accept_image_formats or get_default_accept_image_formats()
        )

    @property
    def pip_dependencies(self):
        return ["imageio"]

    @property
    def config(self):
        return dict(
            super().config,
            accept_image_formats=list(self.accept_image_formats),
            pilmode=self.pilmode,
        )

    def _extract(self, tasks):
        for task in tasks:
            if not all(f is not None for f in task.data):
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} Empty request",
                )
                continue
            try:
                assert all(
                    not getattr(f, "name", None)
                    or check_file_extension(f.name, self.accept_image_formats)
                    for f in task.data
                )
                image_array_tuple = tuple(
                    imageio.imread(f, pilmode=self.pilmode) for f in task.data
                )
                yield image_array_tuple
            except ValueError:
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} "
                    f"Input image decode failed, it must be in supported format list: "
                    f"{self.accept_image_formats}",
                )
            except AssertionError:
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} "
                    f"Input image file must be in supported format list: "
                    f"{self.accept_image_formats}",
                )
            except Exception:  # pylint: disable=broad-except
                err = traceback.format_exc()
                task.discard(
                    http_status=500,
                    err_msg=f"BentoML#{self.__class__.__name__} "
                    f"Internal Server Error: {err}",
                )

    def extract_user_func_args(self, tasks: Iterable[MultiImgTask]) -> ApiFuncArgs:
        args = tuple(map(tuple, zip(*self._extract(tasks))))
        if not args:
            args = (tuple(),) * len(self.input_names)
        return args



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import traceback
from typing import Iterable, Sequence, Tuple

from bentoml.adapters.string_input import StringInput
from bentoml.adapters.utils import TF_B64_KEY
from bentoml.types import InferenceTask, JsonSerializable


def b64_hook(o):
    if isinstance(o, dict) and TF_B64_KEY in o:
        return base64.b64decode(o[TF_B64_KEY])
    return o


ApiFuncArgs = Tuple[
    Sequence[JsonSerializable],
]


class TfTensorInput(StringInput):
    """
    Tensor input adapter for Tensorflow models.
    Transform incoming tf tensor data from http request, cli or lambda event into
    tf tensor.
    The behavior should be compatible with tensorflow serving REST API:
    * https://www.tensorflow.org/tfx/serving/api_rest#classify_and_regress_api
    * https://www.tensorflow.org/tfx/serving/api_rest#predict_api

    Args:
        * method: equivalence of serving API methods: (predict, classify, regress)

    Raises:
        BentoMLException: BentoML currently doesn't support Content-Type

    Examples
    --------
    Example Service:

    .. code-block:: python

        import tensorflow as tf
        import bentoml
        from bentoml.adapters import TfTensorInput
        from bentoml.frameworks.tensorflow import TensorflowSavedModelArtifact

        @bentoml.env(infer_pip_packages=True)
        @bentoml.artifacts([TensorflowSavedModelArtifact('model')])
        class MyService(bentoml.BentoService):

            @bentoml.api(input=TfTensorInput(), batch=True)
            def predict(self, input: tf.Tensor):
                result = self.artifacts.model.predict(input)
                return result

    Query with Http request:

        curl -i \\
        --header "Content-Type: application/json"
        --request POST \\
        --data '{"instances": [1]}' \\
        localhost:5000/predict


    Query with CLI command::

        bentoml run MyService:latest predict --input \\
          '{"instances": [1]}'
    """

    BATCH_MODE_SUPPORTED = True
    SINGLE_MODE_SUPPORTED = False
    METHODS = (PREDICT, CLASSIFY, REGRESS) = ("predict", "classify", "regress")

    def __init__(self, method=PREDICT, **base_kwargs):
        super().__init__(**base_kwargs)
        self.method = method

    @property
    def config(self):
        base_config = super().config
        return dict(base_config, method=self.method,)

    @property
    def request_schema(self):
        if self.method == self.PREDICT:
            return {
                "application/json": {
                    "schema": {
                        "type": "object",
                        "properties": {
                            "signature_name": {"type": "string", "default": None},
                            "instances": {
                                "type": "array",
                                "items": {"type": "object"},
                                "default": None,
                            },
                            "inputs": {"type": "object", "default": None},
                        },
                    }
                }
            }
        else:
            raise NotImplementedError(f"method {self.method} is not implemented")

    def extract_user_func_args(
        self, tasks: Iterable[InferenceTask[str]]
    ) -> ApiFuncArgs:
        import tensorflow as tf

        instances_list = []
        for task in tasks:
            try:
                parsed_json = json.loads(task.data, object_hook=b64_hook)
                if parsed_json.get("instances") is None:
                    task.discard(
                        http_status=400, err_msg="input format is not implemented",
                    )
                else:
                    instances = parsed_json.get("instances")
                    if (
                        task.http_headers.is_batch_input
                        or task.http_headers.is_batch_input is None
                    ):
                        task.batch = len(instances)
                        instances_list.extend(instances)
                    else:
                        instances_list.append(instances)
            except json.JSONDecodeError:
                task.discard(http_status=400, err_msg="Not a valid JSON format")
            except Exception:  # pylint: disable=broad-except
                err = traceback.format_exc()
                task.discard(http_status=500, err_msg=f"Internal Server Error: {err}")

        parsed_tensor = tf.constant(instances_list)
        return (parsed_tensor,)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import logging
from pathlib import Path

from bentoml.exceptions import YataiRepositoryException
from bentoml.yatai.proto.repository_pb2 import BentoUri
from bentoml.yatai.repository.base_repository import BaseRepository


logger = logging.getLogger(__name__)


class FileSystemRepository(BaseRepository):
    def __init__(self, base_url):
        """
        :param base_url: local file system path that will be used as the root directory
            of this saved bundle repository
        """
        if not os.path.exists(base_url):
            # make sure local repo base path exist
            os.mkdir(base_url)

        self.base_path = base_url
        self.uri_type = BentoUri.LOCAL

    def add(self, bento_name, bento_version):
        # Full path containing saved BentoService bundle, it the base path with service
        # name and service version as prefix. e.g.:
        # with base_path = '/tmp/my_bento_repo/', the saved bento will resolve in
        # the directory: '/tmp/my_bento_repo/service_name/version/'
        target_dir = os.path.join(self.base_path, bento_name, bento_version)

        # Ensure parent directory exist
        Path(os.path.join(self.base_path), bento_name).mkdir(
            parents=True, exist_ok=True
        )

        # Raise if target bento version already exist in storage
        if os.path.exists(target_dir):
            raise YataiRepositoryException(
                "Existing BentoService bundle {name}:{version} found in repository: "
                "{target_dir}".format(
                    name=bento_name, version=bento_version, target_dir=target_dir
                )
            )

        return BentoUri(type=self.uri_type, uri=target_dir)

    def get(self, bento_name, bento_version):
        saved_path = os.path.join(self.base_path, bento_name, bento_version)
        if not os.path.exists(saved_path):
            logger.warning(
                "BentoML bundle %s:%s not found in target repository",
                bento_name,
                bento_version,
            )
        return saved_path

    def dangerously_delete(self, bento_name, bento_version):
        saved_path = os.path.join(self.base_path, bento_name, bento_version)
        try:
            return shutil.rmtree(saved_path)
        except FileNotFoundError:
            logger.warning(
                "BentoService %s:%s has already been deleted from local storage",
                bento_name,
                bento_version,
            )
            return



import pickle
from functools import lru_cache
from typing import Sequence

from bentoml.types import HTTPRequest, HTTPResponse

MARSHAL_REQUEST_HEADER = "BentoML-Is-Merged-Request"


class PlasmaDataLoader:
    """
    Transfer datas with plasma store, in development now
    """

    @classmethod
    @lru_cache(maxsize=1)
    def get_plasma(cls):
        import pyarrow.plasma as plasma
        import subprocess

        object_store_size = 2 * 10 * 1000 * 1000
        plasma_path = "/tmp/store"
        subprocess.Popen(
            ["plasma_store", "-s", plasma_path, "-m", str(object_store_size)]
        )
        return plasma.connect(plasma_path)

    @classmethod
    def merge_requests(cls, reqs) -> bytes:
        merged_reqs = tuple((b, h) for h, b in reqs)
        oid = cls.get_plasma().put(merged_reqs)
        return oid.binary()

    @classmethod
    def split_responses(cls, raw: bytes):
        import pyarrow.plasma as plasma

        oid = plasma.ObjectID(raw)
        merged_responses = cls.get_plasma().get(oid)
        cls.get_plasma().delete((oid,))
        return merged_responses

    @classmethod
    def split_requests(cls, raw: bytes):
        import pyarrow.plasma as plasma

        oid = plasma.ObjectID(raw)
        info_list = cls.get_plasma().get(oid)
        cls.get_plasma().delete((oid,))
        return info_list

    @classmethod
    def merge_responses(cls, resps) -> bytes:
        merged_resps = tuple((r, tuple()) for r in resps)
        oid = cls.get_plasma().put(merged_resps)
        return oid.binary()


class PickleDataLoader:
    @classmethod
    def merge_requests(cls, reqs: Sequence[HTTPRequest]) -> bytes:
        return pickle.dumps(reqs)

    @classmethod
    def split_requests(cls, raw: bytes) -> Sequence[HTTPRequest]:
        return pickle.loads(raw)

    @classmethod
    def merge_responses(cls, resps: Sequence[HTTPResponse]) -> bytes:
        return pickle.dumps(list(resps))

    @classmethod
    def split_responses(cls, raw: bytes) -> Sequence[HTTPResponse]:
        try:
            return pickle.loads(raw)
        except pickle.UnpicklingError:
            raise ValueError(
                f"Batching result unpacking error: \n {raw[:1000]}"
            ) from None


DataLoader = PickleDataLoader



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import time
import traceback
from typing import Optional, TYPE_CHECKING

import psutil
from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import RemoteException
from bentoml.marshal.dispatcher import CorkDispatcher, NonBlockSema
from bentoml.marshal.utils import DataLoader, MARSHAL_REQUEST_HEADER
from bentoml.saved_bundle import load_bento_service_metadata
from bentoml.saved_bundle.config import DEFAULT_MAX_BATCH_SIZE, DEFAULT_MAX_LATENCY
from bentoml.tracing import get_tracer
from bentoml.types import HTTPRequest, HTTPResponse


logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from aiohttp_cors import ResourceOptions
    from aiohttp.web import Application, Request
    from aiohttp import BaseConnector, ClientSession


def metrics_patch(cls):
    class _MarshalService(cls):
        @inject
        def __init__(
            self,
            *args,
            namespace: str = Provide[
                BentoMLContainer.config.bento_server.metrics.namespace
            ],
            **kwargs,
        ):
            for attr_name in functools.WRAPPER_ASSIGNMENTS:
                try:
                    setattr(self.__class__, attr_name, getattr(cls, attr_name))
                except AttributeError:
                    pass

            from prometheus_client import Counter, Gauge, Histogram

            super(_MarshalService, self).__init__(*args, **kwargs)
            # its own namespace?
            service_name = self.bento_service_metadata_pb.name

            self.metrics_request_batch_size = Histogram(
                name=service_name + '_mb_batch_size',
                documentation=service_name + "microbatch request batch size",
                namespace=namespace,
                labelnames=['endpoint'],
            )
            self.metrics_request_duration = Histogram(
                name=service_name + '_mb_request_duration_seconds',
                documentation=service_name + "API HTTP request duration in seconds",
                namespace=namespace,
                labelnames=['endpoint', 'http_response_code'],
            )
            self.metrics_request_in_progress = Gauge(
                name=service_name + "_mb_request_in_progress",
                documentation='Total number of HTTP requests in progress now',
                namespace=namespace,
                labelnames=['endpoint', 'http_method'],
            )
            self.metrics_request_exception = Counter(
                name=service_name + "_mb_request_exception",
                documentation='Total number of service exceptions',
                namespace=namespace,
                labelnames=['endpoint', 'exception_class'],
            )
            self.metrics_request_total = Counter(
                name=service_name + "_mb_request_total",
                documentation='Total number of service exceptions',
                namespace=namespace,
                labelnames=['endpoint', 'http_response_code'],
            )

        async def request_dispatcher(self, request):
            from aiohttp.web import Response

            func = super(_MarshalService, self).request_dispatcher
            api_route = request.match_info.get("path", "/")
            _metrics_request_in_progress = self.metrics_request_in_progress.labels(
                endpoint=api_route, http_method=request.method,
            )
            _metrics_request_in_progress.inc()
            time_st = time.time()
            try:
                resp = await func(request)
            except asyncio.CancelledError:
                resp = Response(status=503)
            except Exception as e:  # pylint: disable=broad-except
                self.metrics_request_exception.labels(
                    endpoint=api_route, exception_class=e.__class__.__name__
                ).inc()
                logger.error(traceback.format_exc())
                resp = Response(status=500)
            self.metrics_request_total.labels(
                endpoint=api_route, http_response_code=resp.status
            ).inc()
            self.metrics_request_duration.labels(
                endpoint=api_route, http_response_code=resp.status
            ).observe(time.time() - time_st)
            _metrics_request_in_progress.dec()
            return resp

        async def _batch_handler_template(self, requests, api_route, max_latency):
            func = super(_MarshalService, self)._batch_handler_template
            self.metrics_request_batch_size.labels(endpoint=api_route).observe(
                len(requests)
            )
            return await func(requests, api_route, max_latency)

    return _MarshalService


@metrics_patch
class MarshalService:
    """
    MarshalService creates a reverse proxy server in front of actual API server,
    implementing the micro batching feature.
    It wait a short period and packed multiple requests in a single batch
    before sending to the API server.
    It applied an optimized CORK algorithm to get best efficiency.
    """

    @inject
    def __init__(
        self,
        bento_bundle_path,
        outbound_host="localhost",
        outbound_port=None,
        outbound_workers: int = Provide[BentoMLContainer.api_server_workers],
        mb_max_batch_size: int = Provide[
            BentoMLContainer.config.bento_server.microbatch.max_batch_size
        ],
        mb_max_latency: int = Provide[
            BentoMLContainer.config.bento_server.microbatch.max_latency
        ],
        max_request_size: int = Provide[
            BentoMLContainer.config.bento_server.max_request_size
        ],
        outbound_unix_socket: str = None,
        enable_access_control: bool = Provide[
            BentoMLContainer.config.bento_server.cors.enabled
        ],
        access_control_allow_origin: Optional[str] = Provide[
            BentoMLContainer.config.bento_server.cors.access_control_allow_origin
        ],
        access_control_options: Optional["ResourceOptions"] = Provide[
            BentoMLContainer.access_control_options
        ],
        timeout: int = Provide[BentoMLContainer.config.bento_server.timeout],
    ):

        self._conn: Optional["BaseConnector"] = None
        self._client: Optional["ClientSession"] = None
        self.outbound_unix_socket = outbound_unix_socket
        self.outbound_host = outbound_host
        self.outbound_port = outbound_port
        self.outbound_workers = outbound_workers
        self.mb_max_batch_size = mb_max_batch_size
        self.mb_max_latency = mb_max_latency
        self.batch_handlers = dict()
        self._outbound_sema = None  # the semaphore to limit outbound connections
        self._cleanup_tasks = None
        self.max_request_size = max_request_size

        self.enable_access_control = enable_access_control
        self.access_control_allow_origin = access_control_allow_origin
        self.access_control_options = access_control_options

        self.bento_service_metadata_pb = load_bento_service_metadata(bento_bundle_path)

        self.setup_routes_from_pb(self.bento_service_metadata_pb)
        self.timeout = timeout

        if psutil.POSIX:
            import resource

            self.CONNECTION_LIMIT = resource.getrlimit(resource.RLIMIT_NOFILE)[0]
        else:
            self.CONNECTION_LIMIT = 1024
        logger.info(
            "Your system nofile limit is %d, which means each instance of microbatch "
            "service is able to hold this number of connections at same time. "
            "You can increase the number of file descriptors for the server process, "
            "or launch more microbatch instances to accept more concurrent connection.",
            self.CONNECTION_LIMIT,
        )

    @property
    def cleanup_tasks(self):
        if self._cleanup_tasks is None:
            self._cleanup_tasks = []
        return self._cleanup_tasks

    async def cleanup(self, _):
        # clean up futures for gracefully shutting down
        for task in self.cleanup_tasks:
            await task()
        if getattr(self, '_client', None) is not None and not self._client.closed:
            await self._client.close()

    def set_outbound_port(self, outbound_port):
        self.outbound_port = outbound_port

    def fetch_sema(self):
        if self._outbound_sema is None:
            self._outbound_sema = NonBlockSema(self.outbound_workers)
        return self._outbound_sema

    def get_conn(self) -> "BaseConnector":
        import aiohttp

        if self._conn is None or self._conn.closed:
            if self.outbound_unix_socket:
                self._conn = aiohttp.UnixConnector(path=self.outbound_unix_socket,)
            else:
                self._conn = aiohttp.TCPConnector(limit=30)
        return self._conn

    def get_client(self):
        import aiohttp

        if self._client is None or self._client.closed:
            jar = aiohttp.DummyCookieJar()
            if self.timeout:
                timeout = aiohttp.ClientTimeout(total=self.timeout)
            else:
                timeout = None
            self._client = aiohttp.ClientSession(
                connector=self.get_conn(),
                auto_decompress=False,
                cookie_jar=jar,
                connector_owner=False,
                timeout=timeout,
            )
        return self._client

    def add_batch_handler(self, api_route, max_latency, max_batch_size):
        '''
        Params:
        * max_latency: limit the max latency of overall request handling
        * max_batch_size: limit the max batch size for handler

        ** marshal server will give priority to meet these limits than efficiency
        '''
        from aiohttp.web import HTTPTooManyRequests

        if api_route not in self.batch_handlers:
            dispatcher = CorkDispatcher(
                max_latency,
                max_batch_size,
                shared_sema=self.fetch_sema(),
                fallback=HTTPTooManyRequests,
            )
            _func = dispatcher(
                functools.partial(
                    self._batch_handler_template,
                    api_route=api_route,
                    max_latency=max_latency,
                )
            )
            self.cleanup_tasks.append(dispatcher.shutdown)
            self.batch_handlers[api_route] = _func

    def setup_routes_from_pb(self, bento_service_metadata_pb):
        for api_pb in bento_service_metadata_pb.apis:
            if api_pb.batch:
                max_latency = (
                    self.mb_max_latency or api_pb.mb_max_latency or DEFAULT_MAX_LATENCY
                )
                max_batch_size = (
                    self.mb_max_batch_size
                    or api_pb.mb_max_batch_size
                    or DEFAULT_MAX_BATCH_SIZE
                )
                self.add_batch_handler(api_pb.route, max_latency, max_batch_size)
                logger.info(
                    "Micro batch enabled for API `%s` max-latency: %s"
                    " max-batch-size %s",
                    api_pb.route,
                    max_latency,
                    max_batch_size,
                )

    async def request_dispatcher(self, request: "Request"):
        from aiohttp.web import HTTPInternalServerError, Response

        with get_tracer().async_span(
            service_name=self.__class__.__name__,
            span_name="[1]http request",
            is_root=True,
            standalone=True,
            sample_rate=0.001,
        ):
            api_route = request.match_info.get("path")
            if api_route in self.batch_handlers:
                req = HTTPRequest(
                    tuple((k.decode(), v.decode()) for k, v in request.raw_headers),
                    await request.read(),
                )
                try:
                    resp = await self.batch_handlers[api_route](req)
                except RemoteException as e:
                    # known remote exception
                    logger.error(traceback.format_exc())
                    resp = Response(
                        status=e.payload.status,
                        headers=e.payload.headers,
                        body=e.payload.body,
                    )
                except Exception:  # pylint: disable=broad-except
                    logger.error(traceback.format_exc())
                    resp = HTTPInternalServerError()
            else:
                resp = await self.relay_handler(request)
        return resp

    async def relay_handler(self, request):
        from aiohttp.client_exceptions import ClientConnectionError
        from aiohttp.web import Response

        data = await request.read()
        url = request.url.with_host(self.outbound_host).with_port(self.outbound_port)

        with get_tracer().async_span(
            service_name=self.__class__.__name__,
            span_name=f"[2]{url.path} relay",
            request_headers=request.headers,
        ):
            try:
                client = self.get_client()
                async with client.request(
                    request.method, url, data=data, headers=request.headers
                ) as resp:
                    body = await resp.read()
            except ClientConnectionError:
                return Response(status=503, body=b"Service Unavailable")
        return Response(status=resp.status, body=body, headers=resp.headers,)

    async def _batch_handler_template(self, requests, api_route, max_latency):
        '''
        batch request handler
        params:
            * requests: list of aiohttp request
            * api_route: called API name
        raise:
            * RemoteException: known exceptions from model server
            * Exception: other exceptions
        '''
        from aiohttp.client_exceptions import ClientConnectionError
        from aiohttp import ClientTimeout
        from aiohttp.web import Response

        headers = {MARSHAL_REQUEST_HEADER: "true"}
        api_url = f"http://{self.outbound_host}:{self.outbound_port}/{api_route}"

        with get_tracer().async_span(
            service_name=self.__class__.__name__,
            span_name=f"[2]merged {api_route}",
            request_headers=headers,
        ):
            reqs_s = DataLoader.merge_requests(requests)
            try:
                client = self.get_client()
                timeout = ClientTimeout(
                    total=(self.mb_max_latency or max_latency) // 1000
                )
                async with client.post(
                    api_url, data=reqs_s, headers=headers, timeout=timeout
                ) as resp:
                    raw = await resp.read()
            except ClientConnectionError as e:
                raise RemoteException(
                    e, payload=HTTPResponse.new(status=503, body=b"Service Unavailable")
                )
            except asyncio.CancelledError as e:
                raise RemoteException(
                    e,
                    payload=HTTPResponse(
                        status=500, body=b"Cancelled before upstream responses"
                    ),
                )
            except asyncio.TimeoutError as e:
                raise RemoteException(
                    e, payload=HTTPResponse(status=408, body=b"Request timeout"),
                )

            if resp.status != 200:
                raise RemoteException(
                    f"Bad response status from model server:\n{resp.status}\n{raw}",
                    payload=HTTPResponse.new(
                        status=resp.status,
                        headers=tuple(resp.headers.items()),
                        body=raw,
                    ),
                )
            merged = DataLoader.split_responses(raw)
            return tuple(
                Response(body=i.body, headers=i.headers, status=i.status or 500)
                for i in merged
            )

    def make_app(self) -> "Application":
        from aiohttp.web import Application
        from aiohttp import hdrs

        methods = hdrs.METH_ALL.copy()

        if self.enable_access_control:
            # ref: https://github.com/aio-libs/aiohttp-cors/issues/241
            methods.remove(hdrs.METH_OPTIONS)

        app = Application(client_max_size=self.max_request_size)
        app.on_cleanup.append(self.cleanup)

        for method in methods:
            app.router.add_route(method, "/", self.relay_handler)
            app.router.add_route(method, "/{path:.*}", self.request_dispatcher)

        if self.enable_access_control:
            assert (
                self.access_control_allow_origin is not None
            ), "To enable cors, access_control_allow_origin must be set"
            assert self.access_control_options is not None

            import aiohttp_cors

            # Configure default CORS settings.
            cors = aiohttp_cors.setup(
                app,
                defaults={
                    self.access_control_allow_origin: self.access_control_options
                },
            )
            # Configure CORS on all routes.
            for route in list(app.router.routes()):
                cors.add(route)

        return app

    @inject
    def fork_start_app(
        self, port=Provide[BentoMLContainer.config.bento_server.port],
    ):
        from aiohttp.web import run_app

        # Use new eventloop in the fork process to avoid problems on MacOS
        # ref: https://groups.google.com/forum/#!topic/python-tornado/DkXjSNPCzsI
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        app = self.make_app()
        run_app(app, port=port)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Sequence

from bentoml.adapters.base_output import regroup_return_value
from bentoml.adapters.json_output import JsonOutput
from bentoml.adapters.utils import TfTensorJsonEncoder
from bentoml.types import InferenceError, InferenceResult, InferenceTask
from bentoml.utils.lazy_loader import LazyLoader

np = LazyLoader('np', globals(), 'numpy')


def tf_to_numpy(tensor):
    '''
    Tensor -> ndarray
    List[Tensor] -> tuple[ndarray]
    '''
    import tensorflow as tf

    if isinstance(tensor, (list, tuple)):
        return tuple(tf_to_numpy(t) for t in tensor)

    if tf.__version__.startswith("1."):
        with tf.compat.v1.Session():
            return tensor.numpy()
    else:
        return tensor.numpy()


class TfTensorOutput(JsonOutput):
    """
    Output adapters converts returns of user defined API function into specific output,
    such as HTTP response, command line stdout or AWS Lambda event object.

    Args:
        cors (str): DEPRECATED. Moved to the configuration file.
            The value of the Access-Control-Allow-Origin header set in the
            HTTP/AWS Lambda response object. If set to None, the header will not be set.
            Default is None.
        ensure_ascii(bool): Escape all non-ASCII characters. Default False.
    """

    BATCH_MODE_SUPPORTED = True

    @property
    def pip_dependencies(self):
        """
        :return: List of PyPI package names required by this OutputAdapter
        """
        return ['tensorflow']

    def pack_user_func_return_value(
        self, return_result, tasks: Sequence[InferenceTask],
    ) -> Sequence[InferenceResult[str]]:
        rv = []
        results = tf_to_numpy(return_result)
        for result, _ in regroup_return_value(results, tasks):
            try:
                result_str = json.dumps(result, cls=TfTensorJsonEncoder)
                rv.append(InferenceResult(data=result_str, http_status=200))
            except Exception as e:  # pylint: disable=broad-except
                rv.append(InferenceError(err_msg=str(e), http_status=500))
        return rv



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import multiprocessing
from typing import Optional

from bentoml.utils import reserve_free_port

logger = logging.getLogger(__name__)


def start_dev_server(
    bundle_path: str,
    port: Optional[int] = None,
    mb_max_batch_size: Optional[int] = None,
    mb_max_latency: Optional[int] = None,
    run_with_ngrok: Optional[bool] = None,
    enable_swagger: Optional[bool] = None,
):
    if run_with_ngrok:
        from threading import Timer

        from bentoml.utils.flask_ngrok import start_ngrok

        thread = Timer(1, start_ngrok, args=(port,))
        thread.setDaemon(True)
        thread.start()

    with reserve_free_port() as api_server_port:
        # start server right after port released
        #  to reduce potential race

        model_server_proc = multiprocessing.Process(
            target=_start_dev_server,
            kwargs=dict(
                api_server_port=api_server_port,
                saved_bundle_path=bundle_path,
                enable_swagger=enable_swagger,
            ),
            daemon=True,
        )
    model_server_proc.start()

    try:
        _start_dev_proxy(
            port=port,
            api_server_port=api_server_port,
            saved_bundle_path=bundle_path,
            mb_max_batch_size=mb_max_batch_size,
            mb_max_latency=mb_max_latency,
        )
    finally:
        model_server_proc.terminate()


def _start_dev_server(
    saved_bundle_path: str, api_server_port: int, enable_swagger: bool,
):
    logger.info("Starting BentoML API server in development mode..")

    from bentoml.server.api_server import BentoAPIServer
    from bentoml.saved_bundle import load_from_dir

    bento_service = load_from_dir(saved_bundle_path)
    api_server = BentoAPIServer(bento_service, enable_swagger=enable_swagger)
    api_server.start(port=api_server_port)


def _start_dev_proxy(
    port: int,
    saved_bundle_path: str,
    api_server_port: int,
    mb_max_batch_size: int,
    mb_max_latency: int,
):
    logger.info("Starting BentoML API proxy in development mode..")

    from bentoml.marshal.marshal import MarshalService

    marshal_server = MarshalService(
        saved_bundle_path,
        outbound_host="localhost",
        outbound_port=api_server_port,
        mb_max_batch_size=mb_max_batch_size,
        mb_max_latency=mb_max_latency,
    )

    marshal_server.fork_start_app(port=port)


def start_prod_server(
    saved_bundle_path: str,
    port: Optional[int] = None,
    workers: Optional[int] = None,
    timeout: Optional[int] = None,
    enable_swagger: Optional[bool] = None,
    mb_max_batch_size: Optional[int] = None,
    mb_max_latency: Optional[int] = None,
    microbatch_workers: Optional[int] = None,
):

    import psutil

    assert (
        psutil.POSIX
    ), "BentoML API Server production mode only supports POSIX platforms"

    prometheus_lock = multiprocessing.Lock()
    with reserve_free_port() as api_server_port:
        pass

    model_server_job = multiprocessing.Process(
        target=_start_prod_server,
        kwargs=dict(
            saved_bundle_path=saved_bundle_path,
            port=api_server_port,
            timeout=timeout,
            workers=workers,
            prometheus_lock=prometheus_lock,
            enable_swagger=enable_swagger,
        ),
        daemon=True,
    )
    model_server_job.start()

    try:
        _start_prod_proxy(
            saved_bundle_path=saved_bundle_path,
            port=port,
            api_server_port=api_server_port,
            workers=microbatch_workers,
            timeout=timeout,
            outbound_workers=workers,
            mb_max_batch_size=mb_max_batch_size,
            mb_max_latency=mb_max_latency,
            prometheus_lock=prometheus_lock,
        )
    finally:
        model_server_job.terminate()


def _start_prod_server(
    saved_bundle_path: str,
    port: int,
    timeout: int,
    workers: int,
    enable_swagger: bool,
    prometheus_lock: Optional[multiprocessing.Lock] = None,
):

    logger.info("Starting BentoML API server in production mode..")

    from bentoml.server.gunicorn_server import gunicorn_bento_server

    gunicorn_app = gunicorn_bento_server()(
        saved_bundle_path,
        port=port,
        timeout=timeout,
        workers=workers,
        prometheus_lock=prometheus_lock,
        enable_swagger=enable_swagger,
    )
    gunicorn_app.run()


def _start_prod_proxy(
    saved_bundle_path: str,
    port: int,
    api_server_port: int,
    workers: int,
    timeout: int,
    outbound_workers: int,
    mb_max_batch_size: int,
    mb_max_latency: int,
    prometheus_lock: Optional[multiprocessing.Lock] = None,
):

    logger.info("Starting BentoML proxy in production mode..")

    from bentoml.server.marshal_server import gunicorn_marshal_server

    # avoid load model before gunicorn fork
    marshal_server = gunicorn_marshal_server()(
        bundle_path=saved_bundle_path,
        prometheus_lock=prometheus_lock,
        port=port,
        workers=workers,
        timeout=timeout,
        outbound_host="localhost",
        outbound_port=api_server_port,
        outbound_workers=outbound_workers,
        mb_max_batch_size=mb_max_batch_size,
        mb_max_latency=mb_max_latency,
    )
    marshal_server.run()



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import grpc
from bentoml.utils.lazy_loader import LazyLoader

yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')


def _proto_status_code_to_http_status_code(proto_status_code, fallback):
    _PROTO_STATUS_CODE_TO_HTTP_STATUS_CODE = {
        yatai_proto.status_pb2.Status.INTERNAL: 500,  # Internal Server Error
        yatai_proto.status_pb2.Status.INVALID_ARGUMENT: 400,  # "Bad Request"
        yatai_proto.status_pb2.Status.NOT_FOUND: 404,  # Not Found
        yatai_proto.status_pb2.Status.DEADLINE_EXCEEDED: 408,  # Request Time out
        yatai_proto.status_pb2.Status.PERMISSION_DENIED: 401,  # Unauthorized
        yatai_proto.status_pb2.Status.UNAUTHENTICATED: 401,  # Unauthorized
        yatai_proto.status_pb2.Status.FAILED_PRECONDITION: 500,  # Internal Server Error
    }
    return _PROTO_STATUS_CODE_TO_HTTP_STATUS_CODE.get(proto_status_code, fallback)


class BentoMLException(Exception):
    """
    Base class for all BentoML's errors.
    Each custom exception should be derived from this class
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.INTERNAL

    @property
    def status_proto(self):
        return yatai_proto.status_pb2.Status(
            status_code=self.proto_status_code, error_message=str(self)
        )

    @property
    def status_code(self):
        """HTTP response status code"""
        return _proto_status_code_to_http_status_code(self.proto_status_code, 500)


class RemoteException(BentoMLException):
    """
    Raise when known exceptions happened in remote server(a model server normally)
    """

    def __init__(self, *args, payload, **kwargs):
        super(RemoteException, self).__init__(*args, **kwargs)
        self.payload = payload


class BentoMLRpcError(BentoMLException):
    def __init__(self, grpc_error, message):
        super(BentoMLRpcError, self).__init__()
        self.grpc_error = grpc_error
        self.message = message
        if self.grpc_error.code == grpc.StatusCode.DEADLINE_EXCEEDED:
            self.grpc_error_message = 'Request time out'
        else:
            self.grpc_error_message = self.grpc_error.details()

    def __str__(self):
        return f'{self.message}: {self.grpc_error_message}'


class Unauthenticated(BentoMLException):
    """
    Raise when a BentoML operation is not authenticated properly, either against 3rd
    party cloud service such as AWS s3, Docker Hub, or Atalaya hosted BentoML service
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.UNAUTHENTICATED


class InvalidArgument(BentoMLException):
    """
    Raise when BentoML received unexpected/invalid arguments from CLI arguments, HTTP
    Request, or python API function parameters
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.INVALID_ARGUMENT


class APIDeprecated(BentoMLException):
    """
    Raise when trying to use deprecated APIs of BentoML
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.INVALID_ARGUMENT


class BadInput(InvalidArgument):
    """Raise when InputAdapter receiving bad input request"""


class NotFound(BentoMLException):
    """
    Raise when specified resource or name not found
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.NOT_FOUND


class FailedPrecondition(BentoMLException):
    """
    Raise when required precondition check failed
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.FAILED_PRECONDITION


class LockUnavailable(BentoMLException):
    """
    Raise when a bundle/deployment resource is unable to be locked
    """

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.FAILED_PRECONDITION


class ArtifactLoadingException(BentoMLException):
    """Raise when BentoService failed to load model artifacts from saved bundle"""


class BentoMLConfigException(BentoMLException):
    """Raise when BentoML is misconfigured or when required configuration is missing"""


class MissingDependencyException(BentoMLException):
    """
    Raise when BentoML component failed to load required dependency - some BentoML
    components has dependency that is optional to the library itself. For example,
    when using SklearnModelArtifact, the scikit-learn module is required although
    BentoML does not require scikit-learn to be a dependency when installed
    """


class YataiServiceException(BentoMLException):
    """Raise when YataiService encounters an error"""


class YataiServiceRpcAborted(YataiServiceException):
    """Raise when YataiService RPC operation aborted"""

    @property
    def proto_status_code(self):
        return yatai_proto.status_pb2.Status.ABORTED


class YataiDeploymentException(YataiServiceException):
    """Raise when YataiService encounters an issue creating/managing deployments"""


class YataiRepositoryException(YataiServiceException):
    """Raise when YataiService encounters an issue managing BentoService repository"""


class AWSServiceError(YataiDeploymentException):
    """Raise when YataiService encounters an issue with AWS service"""


class AzureServiceError(YataiDeploymentException):
    """Raise when YataiService encounters an issue with Azure service"""


class CLIException(BentoMLException):
    """Raise when CLI encounters an issue"""


class YataiLabelException(YataiServiceException):
    """Raise when YataiService encounters an issue managing BentoService label"""



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod, ABCMeta

from bentoml.yatai.proto.deployment_pb2 import DeploymentSpec
from bentoml.exceptions import YataiDeploymentException


def get_deployment_operator(yatai_service, deployment_pb):
    operator = deployment_pb.spec.operator

    if operator == DeploymentSpec.AWS_SAGEMAKER:
        from bentoml.yatai.deployment.sagemaker.operator import (
            SageMakerDeploymentOperator,
        )

        return SageMakerDeploymentOperator(yatai_service)
    elif operator == DeploymentSpec.AWS_LAMBDA:
        from bentoml.yatai.deployment.aws_lambda.operator import (
            AwsLambdaDeploymentOperator,
        )

        return AwsLambdaDeploymentOperator(yatai_service)
    elif operator == DeploymentSpec.AZURE_FUNCTIONS:
        from bentoml.yatai.deployment.azure_functions.operator import (
            AzureFunctionsDeploymentOperator,
        )

        return AzureFunctionsDeploymentOperator(yatai_service)
    elif operator == DeploymentSpec.AWS_EC2:
        from bentoml.yatai.deployment.aws_ec2.operator import AwsEc2DeploymentOperator

        return AwsEc2DeploymentOperator(yatai_service)
    elif operator == DeploymentSpec.CUSTOM:
        raise NotImplementedError(
            "Custom deployment operator is not supported in current version of BentoML"
        )
    else:
        raise YataiDeploymentException("DeployOperator must be set")


class DeploymentOperatorBase(object):
    def __init__(self, yatai_service):
        self.yatai_service = yatai_service

    __metaclass__ = ABCMeta

    @abstractmethod
    def add(self, deployment_pb):
        """
        Create deployment based on deployment_pb spec - the bento name and version
        must be found in the given BentoRepository
        """

    @abstractmethod
    def update(self, deployment_pb, previous_deployment):
        """
        Update existing deployment based on deployment_pb spec
        """

    @abstractmethod
    def delete(self, deployment_pb):
        """
        Delete deployment based on deployment_pb spec
        """

    @abstractmethod
    def describe(self, deployment_pb):
        """
        Fetch current state of an existing deployment created with deployment_pb spec
        """



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from urllib.parse import urlparse

from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import YataiRepositoryException
from bentoml.yatai.proto.repository_pb2 import BentoUri
from bentoml.yatai.repository.base_repository import BaseRepository

logger = logging.getLogger(__name__)


class GCSRepository(BaseRepository):
    @inject
    def __init__(
        self,
        base_url,
        expiration: int = Provide[
            BentoMLContainer.config.yatai.repository.gcs.expiration
        ],
    ):
        try:
            from google.cloud import storage
        except ImportError:
            raise YataiRepositoryException(
                '"google-cloud-storage" package is required for Google Cloud '
                'Storage Repository. You can install it with pip: '
                '"pip install google-cloud-storage"'
            )
        self.uri_type = BentoUri.GCS

        parse_result = urlparse(base_url)
        self.bucket = parse_result.netloc
        self.base_path = parse_result.path.lstrip('/')
        self.gcs_client = storage.Client()
        self.expiration = expiration

    def _get_object_name(self, bento_name, bento_version):
        if self.base_path:
            return "/".join([self.base_path, bento_name, bento_version]) + '.tar.gz'
        else:
            return "/".join([bento_name, bento_version]) + '.tar.gz'

    def add(self, bento_name, bento_version):
        object_name = self._get_object_name(bento_name, bento_version)
        try:
            bucket = self.gcs_client.bucket(self.bucket)
            blob = bucket.blob(object_name)

            response = blob.generate_signed_url(
                version="v4", expiration=self.expiration, method="PUT",
            )
        except Exception as e:
            raise YataiRepositoryException(
                "Not able to get pre-signed URL on GCS. Error: {}".format(e)
            )

        return BentoUri(
            type=self.uri_type,
            uri='gs://{}/{}'.format(self.bucket, object_name),
            gcs_presigned_url=response,
        )

    def get(self, bento_name, bento_version):
        # Return gcs path containing uploaded Bento files

        object_name = self._get_object_name(bento_name, bento_version)

        try:
            bucket = self.gcs_client.bucket(self.bucket)
            blob = bucket.blob(object_name)

            response = blob.generate_signed_url(
                version="v4", expiration=self.expiration, method="GET",
            )
            return response
        except Exception:  # pylint: disable=broad-except
            logger.error(
                "Failed generating presigned URL for downloading saved bundle from GCS,"
                "falling back to using gs path and client side credential for"
                "downloading with google.cloud.storage"
            )
            return 'gs://{}/{}'.format(self.bucket, object_name)

    def dangerously_delete(self, bento_name, bento_version):
        # Remove gcs path containing related Bento files

        object_name = self._get_object_name(bento_name, bento_version)

        try:
            bucket = self.gcs_client.bucket(self.bucket)
            blob = bucket.blob(object_name)
            blob.delete()
        except Exception as e:
            raise YataiRepositoryException(
                "Not able to delete object on GCS. Error: {}".format(e)
            )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Iterator, Sequence, Tuple

from bentoml.adapters.base_input import BaseInputAdapter, parse_cli_inputs
from bentoml.adapters.utils import decompress_gzip_request
from bentoml.types import AwsLambdaEvent, FileLike, HTTPRequest, InferenceTask

ApiFuncArgs = Tuple[Sequence[FileLike], ...]
MultiFileTask = InferenceTask[Tuple[FileLike, ...]]


class MultiFileInput(BaseInputAdapter):
    """ Low level input adapters that transform incoming files data from http request,
    CLI or AWS lambda event into binary stream objects, then pass down to user defined
    API functions.

    Parameters
    ----------
    input_names : List[str]
        list of input names. For HTTP they are form input names. For CLI
        they are CLI args --input-<name1> or --input-file-<name1>

    allow_none : bool
        accept HTTP requests or AWS Lambda events without all files
        provided. Does not take effect on CLI.

    Examples
    ----------
    Service using MultiFileInput:

    .. code-block:: python

        from typing import List

        from PIL import Image
        import numpy as np
        import bentoml
        from bentoml.types import FileLike
        from bentoml.framework.pytroch import PytorchModelArtifact
        from bentoml.adapters import MultiFileInput

        @bentoml.env(pip_packages=['torch', 'pillow', 'numpy'])
        @bentoml.artifacts([PytorchModelArtifact('classifier')])
        class PyTorchFashionClassifier(bentoml.BentoService):
            @bentoml.api(
                input=MultiFileInput(input_names=['image', 'json']), batch=True)
            def predict(self, image_list: List[FileLike], json_list: List[FileLike]):
                inputs = []
                for img_io, json_io in zip(image_list, json_list):
                    img = Image.open(img_io)
                    json_obj = json.load(json_io)
                    inputs.append([img, json_obj])
                outputs = self.artifacts.classifier(inputs)
                return outputs

    Query with HTTP request performed by cURL::

        curl -i \\
          -F image=@test.jpg \\
          -F json=@test.json \\
          localhost:5000/predict

    OR by an HTML form that sends multipart data:

    .. code-block:: html

        <form action="http://localhost:8000" method="POST"
              enctype="multipart/form-data">
            <input name="image" type="file">
            <input name="json" type="file">
            <input type="submit">
        </form>

    OR by python requests:

    .. code-block:: python

        import requests

        with open("test.jpg", "rb") as f:
            image_bytes = f.read()
        with open("anno.json", "rb") as f:
            json_bytes = f.read()

        files = {
            "image": ("test.jpg", image_bytes),
            "json": ("test.json", json_bytes),
        }
        response = requests.post(your_url, files=files)

    Query with CLI command::

        bentoml run PyTorchFashionClassifier:latest predict \\
          --input-file-image test.jpg \\
          --input-file-json test.json

    OR infer all file pairs under a folder with ten pairs each batch::

        bentoml run PyTorchFashionClassifier:latest predict --max-batch-size 10 \\
          --input-file-image folder/*.jpg \\
          --input-file-json folder/*.json

    Note: jpg files and json files should be in same prefix like this::

        folder:
            - apple.jpg
            - apple.json
            - banana.jpg
            - banana.json
            ...

    """

    HTTP_METHODS = ["POST"]
    BATCH_MODE_SUPPORTED = True

    def __init__(
        self, input_names: Sequence[str], allow_none: bool = False, **base_kwargs,
    ):

        super().__init__(**base_kwargs)
        self.input_names = input_names
        self.allow_none = allow_none

    @property
    def config(self):
        return {
            # Converting to list, google.protobuf.Struct does not work with tuple type
            "input_names": list(self.input_names)
        }

    @property
    def request_schema(self):
        return {
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {
                        k: {"type": "string", "format": "binary"}
                        for k in self.input_names
                    },
                }
            },
        }

    @decompress_gzip_request
    def from_http_request(self, req: HTTPRequest) -> MultiFileTask:
        if req.headers.content_type != 'multipart/form-data':
            task = InferenceTask(data=None)
            task.discard(
                http_status=400,
                err_msg=f"BentoML#{self.__class__.__name__} only accepts requests "
                "with Content-Type: multipart/form-data",
            )
        else:
            _, _, files = HTTPRequest.parse_form_data(req)
            files = tuple(files.get(k) for k in self.input_names)
            if not any(files):
                task = InferenceTask(data=None)
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} requires inputs "
                    f"fields {self.input_names}",
                )
            elif not all(files) and not self.allow_none:
                task = InferenceTask(data=None)
                task.discard(
                    http_status=400,
                    err_msg=f"BentoML#{self.__class__.__name__} requires inputs "
                    f"fields {self.input_names}",
                )
            else:
                task = InferenceTask(http_headers=req.headers, data=files,)
        return task

    def from_aws_lambda_event(self, event: AwsLambdaEvent) -> MultiFileTask:
        request = HTTPRequest(
            headers=tuple((k, v) for k, v in event.get('headers', {}).items()),
            body=event['body'],
        )
        return self.from_http_request(request)

    def from_cli(self, cli_args: Sequence[str]) -> Iterator[MultiFileTask]:
        input_, input_file = parse_cli_inputs(cli_args, self.input_names)
        for t in self.from_inference_job(input_=input_, input_file=input_file):
            t.cli_args = cli_args
            yield t

    def from_inference_job(  # pylint: disable=arguments-differ
        self, input_=None, input_file=None, **extra_args
    ) -> Iterator[InferenceTask[FileLike]]:
        if input_file is not None:
            for ds in zip(*input_file):
                uris = (pathlib.Path(d).absolute().as_uri() for d in ds)
                fs = tuple(FileLike(uri=uri) for uri in uris)
                yield InferenceTask(data=fs, inference_job_args=extra_args)
        else:
            for ds in zip(*input_):
                fs = tuple(FileLike(bytes_=d.encode()) for d in ds)
                yield InferenceTask(data=fs, inference_job_args=extra_args)

    def extract_user_func_args(self, tasks: Sequence[MultiFileTask]) -> ApiFuncArgs:
        args = tuple(map(tuple, zip(*map(lambda t: t.data, tasks))))
        if not args:
            args = (tuple(),) * len(self.input_names)
        return args
